# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time

import grpc
import sample_application_pb2_grpc

from newrelic.api.transaction import current_transaction

# This import format is to resolve a bug within protobuf 4
# Issues for reference:
# https://github.com/protocolbuffers/protobuf/issues/10075
# https://github.com/protocolbuffers/protobuf/issues/10151
# Within sample_application_pb2.py, the protobuf import can only
# be done once before the DESCRIPTOR value is set to None
# (in subsequent imports) instead of overriding/ignoring the imports.
# This ensures that the imports happen once.
Message = sample_application_pb2_grpc.sample__application__pb2.Message
add_SampleApplicationServicer_to_server = sample_application_pb2_grpc.add_SampleApplicationServicer_to_server
SampleApplicationStub = sample_application_pb2_grpc.SampleApplicationStub


class Status(object):
    code = grpc.StatusCode.ABORTED
    details = "abort_with_status"
    trailing_metadata = {}


class SampleApplicationServicer(sample_application_pb2_grpc.SampleApplicationServicer):
    def DoUnaryUnary(self, request, context):
        context.set_trailing_metadata([("content-type", "text/plain")])
        if request.timesout:
            while context.is_active():
                time.sleep(0.1)
        return Message(text="unary_unary: %s" % request.text)

    def DoUnaryStream(self, request, context):
        context.set_trailing_metadata([("content-type", "text/plain")])
        if request.timesout:
            while context.is_active():
                time.sleep(0.1)
        for i in range(request.count):
            yield Message(text="unary_stream: %s" % request.text)

    def DoStreamUnary(self, request_iter, context):
        context.set_trailing_metadata([("content-type", "text/plain")])
        for request in request_iter:
            if request.timesout:
                while context.is_active():
                    time.sleep(0.1)
            return Message(text="stream_unary: %s" % request.text)

    def DoStreamStream(self, request_iter, context):
        context.set_trailing_metadata([("content-type", "text/plain")])
        for request in request_iter:
            if request.timesout:
                while context.is_active():
                    time.sleep(0.1)
            yield Message(text="stream_stream: %s" % request.text)

    def DoUnaryUnaryRaises(self, request, context):
        raise AssertionError("unary_unary: %s" % request.text)

    def DoUnaryStreamRaises(self, request, context):
        raise AssertionError("unary_stream: %s" % request.text)

    def DoStreamUnaryRaises(self, request_iter, context):
        for request in request_iter:
            raise AssertionError("stream_unary: %s" % request.text)

    def DoStreamStreamRaises(self, request_iter, context):
        for request in request_iter:
            raise AssertionError("stream_stream: %s" % request.text)

    def NoTxnUnaryUnaryRaises(self, request, context):
        current_transaction().ignore_transaction = True
        raise AssertionError("unary_unary: %s" % request.text)

    def NoTxnUnaryStreamRaises(self, request, context):
        current_transaction().ignore_transaction = True
        raise AssertionError("unary_stream: %s" % request.text)

    def NoTxnStreamUnaryRaises(self, request_iter, context):
        current_transaction().ignore_transaction = True
        for request in request_iter:
            raise AssertionError("stream_unary: %s" % request.text)

    def NoTxnStreamStreamRaises(self, request_iter, context):
        current_transaction().ignore_transaction = True
        for request in request_iter:
            raise AssertionError("stream_stream: %s" % request.text)

    def NoTxnUnaryUnary(self, request, context):
        current_transaction().ignore_transaction = True
        return self.DoUnaryUnary(request, context)

    def NoTxnUnaryStream(self, request, context):
        current_transaction().ignore_transaction = True
        return self.DoUnaryStream(request, context)

    def NoTxnStreamUnary(self, request_iter, context):
        current_transaction().ignore_transaction = True
        return self.DoStreamUnary(request_iter, context)

    def NoTxnStreamStream(self, request_iter, context):
        current_transaction().ignore_transaction = True
        return self.DoStreamStream(request_iter, context)

    def DoUnaryUnaryAbort(self, request, context):
        context.abort(grpc.StatusCode.ABORTED, "aborting")

    def DoUnaryStreamAbort(self, request, context):
        context.abort(grpc.StatusCode.ABORTED, "aborting")

    def DoStreamUnaryAbort(self, request_iter, context):
        context.abort(grpc.StatusCode.ABORTED, "aborting")

    def DoStreamStreamAbort(self, request_iter, context):
        context.abort(grpc.StatusCode.ABORTED, "aborting")

    def DoUnaryUnaryAbortWithStatus(self, request, context):
        context.abort_with_status(Status)

    def DoUnaryStreamAbortWithStatus(self, request, context):
        context.abort_with_status(Status)

    def DoStreamUnaryAbortWithStatus(self, request_iter, context):
        context.abort_with_status(Status)

    def DoStreamStreamAbortWithStatus(self, request_iter, context):
        context.abort_with_status(Status)

    def extract_dt_value(self, metadata):
        d = {k: v for k, v in metadata}
        return Message(text=json.dumps(d))

    def DtNoTxnUnaryUnary(self, request, context):
        current_transaction().ignore_transaction = True
        return self.extract_dt_value(context.invocation_metadata())

    def DtNoTxnUnaryStream(self, request, context):
        current_transaction().ignore_transaction = True
        yield self.extract_dt_value(context.invocation_metadata())

    def DtNoTxnStreamUnary(self, request_iter, context):
        current_transaction().ignore_transaction = True
        list(request_iter)  # consume iterator
        return self.extract_dt_value(context.invocation_metadata())

    def DtNoTxnStreamStream(self, request_iter, context):
        current_transaction().ignore_transaction = True
        for request in request_iter:
            yield self.extract_dt_value(context.invocation_metadata())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.core.database_utils import SQLConnections


def validate_transaction_slow_sql_count(num_slow_sql):
    @transient_function_wrapper(
        "newrelic.core.stats_engine", "StatsEngine.record_transaction"
    )
    def _validate_transaction_slow_sql_count(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)
        connections = SQLConnections()

        with connections:
            slow_sql_traces = instance.slow_sql_data(connections)
            assert len(slow_sql_traces) == num_slow_sql

        return result

    return _validate_transaction_slow_sql_count



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pylibmc

from testing_support.db_settings import memcached_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

from newrelic.api.background_task import background_task
from newrelic.api.transaction import set_background_task


DB_SETTINGS = memcached_settings()[0]

MEMCACHED_HOST = DB_SETTINGS["host"]
MEMCACHED_PORT = DB_SETTINGS["port"]
MEMCACHED_NAMESPACE = DB_SETTINGS["namespace"]

MEMCACHED_ADDR = '%s:%s' % (MEMCACHED_HOST, MEMCACHED_PORT)

_test_bt_set_get_delete_scoped_metrics = [
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

_test_bt_set_get_delete_rollup_metrics = [
        ('Datastore/all', 3),
        ('Datastore/allOther', 3),
        ('Datastore/Memcached/all', 3),
        ('Datastore/Memcached/allOther', 3),
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

@validate_transaction_metrics(
        'test_memcache:test_bt_set_get_delete',
        scoped_metrics=_test_bt_set_get_delete_scoped_metrics,
        rollup_metrics=_test_bt_set_get_delete_rollup_metrics,
        background_task=True)
@background_task()
def test_bt_set_get_delete():
    set_background_task(True)
    client = pylibmc.Client([MEMCACHED_ADDR])

    key = MEMCACHED_NAMESPACE + 'key'

    client.set(key, 'value')
    value = client.get(key)
    client.delete(key)

    assert value == 'value'

_test_wt_set_get_delete_scoped_metrics = [
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

_test_wt_set_get_delete_rollup_metrics = [
        ('Datastore/all', 3),
        ('Datastore/allWeb', 3),
        ('Datastore/Memcached/all', 3),
        ('Datastore/Memcached/allWeb', 3),
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

@validate_transaction_metrics(
        'test_memcache:test_wt_set_get_delete',
        scoped_metrics=_test_wt_set_get_delete_scoped_metrics,
        rollup_metrics=_test_wt_set_get_delete_rollup_metrics,
        background_task=False)
@background_task()
def test_wt_set_get_delete():
    set_background_task(False)
    client = pylibmc.Client([MEMCACHED_ADDR])

    key = MEMCACHED_NAMESPACE + 'key'

    client.set(key, 'value')
    value = client.get(key)
    client.delete(key)

    assert value == 'value'



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections import namedtuple

import pytest
from _target_application import make_app
from aiohttp.test_utils import AioHTTPTestCase
from aiohttp.test_utils import TestClient as _TestClient
from testing_support.fixture.event_loop import (  # noqa: F401 pylint: disable=W0611
    event_loop,
)
from testing_support.fixtures import (  # noqa: F401 pylint: disable=W0611
    code_coverage_fixture,
    collector_agent_registration_fixture,
    collector_available_fixture,
)
from testing_support.mock_external_http_server import (
    MockExternalHTTPHResponseHeadersServer,
    MockExternalHTTPServer,
)

_coverage_source = [
    "newrelic.hooks.framework_aiohttp",
]

code_coverage = code_coverage_fixture(source=_coverage_source)

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (framework_aiohttp)", default_settings=_default_settings
)


ServerInfo = namedtuple("ServerInfo", ("base_metric", "url"))


class SimpleAiohttpApp(AioHTTPTestCase):
    def __init__(self, server_cls, middleware, *args, **kwargs):
        super(SimpleAiohttpApp, self).__init__(*args, **kwargs)
        self.server_cls = server_cls
        self.middleware = None
        if middleware:
            self.middleware = [middleware]

    def setUp(self):
        super(SimpleAiohttpApp, self).setUp()
        if hasattr(self, "asyncSetUp"):
            asyncio.get_event_loop().run_until_complete(self.asyncSetUp())
        asyncio.set_event_loop(self.loop)

    def get_app(self, *args, **kwargs):
        return make_app(self.middleware)

    def tearDown(self):
        super(SimpleAiohttpApp, self).tearDown()
        if hasattr(self, "asyncTearDown"):
            asyncio.get_event_loop().run_until_complete(self.asyncTearDown())

    async def _get_client(self, app_or_server):
        """Return a TestClient instance."""
        client_constructor_arg = app_or_server

        scheme = "http"
        host = "127.0.0.1"
        server_kwargs = {}
        if self.server_cls:
            test_server = self.server_cls(app_or_server, scheme=scheme, host=host, **server_kwargs)
            client_constructor_arg = test_server

        return _TestClient(client_constructor_arg)

    get_client = _get_client


@pytest.fixture()
def aiohttp_app(request):
    try:
        middleware = request.getfixturevalue("middleware")
    except:
        middleware = None
    try:
        server_cls = request.getfixturevalue("server_cls")
    except:
        server_cls = None
    case = SimpleAiohttpApp(server_cls=server_cls, middleware=middleware)
    case.setUp()
    yield case
    case.tearDown()


@pytest.fixture(scope="session")
def mock_header_server():
    def handler(self):
        if self.command != "GET":
            self.send_response(501)
            self.end_headers()
            return

        response = str(self.headers).encode("utf-8")
        self.send_response(200)
        self.end_headers()
        self.wfile.write(response)

    with MockExternalHTTPHResponseHeadersServer(handler=handler) as _server:
        yield _server


@pytest.fixture(scope="session")
def mock_external_http_server():
    response_values = []

    def respond_with_cat_header(self):
        headers, response_code = response_values.pop()
        self.send_response(response_code)
        for header, value in headers:
            self.send_header(header, value)
        self.end_headers()
        self.wfile.write(b"")

    with MockExternalHTTPServer(handler=respond_with_cat_header) as server:
        yield (server, response_values)


@pytest.fixture(scope="session")
def local_server_info(mock_header_server):
    host_port = "127.0.0.1:%d" % mock_header_server.port
    metric = "External/%s/aiohttp/" % host_port
    url = "http://" + host_port
    return ServerInfo(metric, url)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import django

from newrelic.core.config import global_settings
from newrelic.common.encoding_utils import gzip_decompress
from testing_support.fixtures import (
    override_application_settings,
    override_generic_settings, override_ignore_status_codes)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

DJANGO_VERSION = tuple(map(int, django.get_version().split('.')[:2]))

if DJANGO_VERSION[0] < 3:
    pytest.skip("support for asgi added in django 3", allow_module_level=True)

from testing_support.asgi_testing import AsgiTest


scoped_metrics = [
    ('Function/django.contrib.sessions.middleware:SessionMiddleware', 1),
    ('Function/django.middleware.common:CommonMiddleware', 1),
    ('Function/django.middleware.csrf:CsrfViewMiddleware', 1),
    ('Function/django.contrib.auth.middleware:AuthenticationMiddleware', 1),
    ('Function/django.contrib.messages.middleware:MessageMiddleware', 1),
    ('Function/django.middleware.gzip:GZipMiddleware', 1),
    ('Function/middleware:ExceptionTo410Middleware', 1),
    ('Function/django.urls.resolvers:URLResolver.resolve', 'present'),
]

rollup_metrics = scoped_metrics + [
    ('Python/Framework/Django/%s' % django.get_version(), 1),
]


@pytest.fixture
def application():
    from django.core.asgi import get_asgi_application

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
    return AsgiTest(get_asgi_application())


@validate_transaction_metrics('views:index',
        scoped_metrics=[('Function/views:index', 1)] + scoped_metrics,
        rollup_metrics=rollup_metrics)
@validate_code_level_metrics("views", "index")
def test_asgi_index(application):
    response = application.get('/')
    assert response.status == 200


@validate_transaction_metrics('views:exception',
        scoped_metrics=[('Function/views:exception', 1)] + scoped_metrics,
        rollup_metrics=rollup_metrics)
@validate_code_level_metrics("views", "exception")
def test_asgi_exception(application):
    response = application.get('/exception')
    assert response.status == 500


@override_ignore_status_codes([410])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:middleware_410',
        scoped_metrics=[('Function/views:middleware_410', 1)] + scoped_metrics,
        rollup_metrics=rollup_metrics)
@validate_code_level_metrics("views", "middleware_410")
def test_asgi_middleware_ignore_status_codes(application):
    response = application.get('/middleware_410')
    assert response.status == 410


@override_ignore_status_codes([403])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:permission_denied',
        scoped_metrics=[('Function/views:permission_denied', 1)] + scoped_metrics,
        rollup_metrics=rollup_metrics)
@validate_code_level_metrics("views", "permission_denied")
def test_asgi_ignored_status_code(application):
    response = application.get('/permission_denied')
    assert response.status == 403


@pytest.mark.parametrize('url,view_name', (
    ('/cbv', 'views:MyView.get'),
    ('/deferred_cbv', 'views:deferred_cbv'),
))
def test_asgi_class_based_view(application, url, view_name):
    func = 'get' if url == '/cbv' else 'deferred_cbv'
    namespace = 'views.MyView' if func == 'get' else 'views'
   
    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(view_name,
            scoped_metrics=[('Function/' + view_name, 1)] + scoped_metrics,
            rollup_metrics=rollup_metrics)
    @validate_code_level_metrics(namespace, func)
    def _test():
        response = application.get(url)
        assert response.status == 200

    _test()


@pytest.mark.parametrize('url', (
    '/html_insertion',
    '/html_insertion_content_length',
    '/template_tags',
    '/gzip_html_insertion',
))
@override_application_settings({
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
})
def test_asgi_html_insertion_success(application, url):
    if 'gzip' in url:
        headers = {'Accept-Encoding': 'gzip'}
    else:
        headers = None
    response = application.get(url, headers=headers)
    assert response.status == 200

    if 'gzip' in url:
        body = gzip_decompress(response.body).encode('utf-8')
    else:
        body = response.body

    assert b'NREUM HEADER' in body
    assert b'NREUM.info' in body
    assert b'&lt;!-- NREUM HEADER --&gt' not in body


@pytest.mark.parametrize('url', (
    '/html_insertion_manual',
    '/html_insertion_unnamed_attachment_header',
    '/html_insertion_named_attachment_header',
))
@override_application_settings({
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
})
def test_asgi_html_insertion_failed(application, url):
    response = application.get(url)
    assert response.status == 200

    assert b'NREUM HEADER' not in response.body
    assert b'NREUM.info' not in response.body


@validate_transaction_metrics('views:template_tags',
        scoped_metrics=[
                ('Function/views:template_tags', 1),
                ('Template/Render/main.html', 1),
                ('Template/Render/results.html', 1)] + scoped_metrics,
        rollup_metrics=rollup_metrics)
@validate_code_level_metrics('views', 'template_tags')
def test_asgi_template_render(application):
    response = application.get('/template_tags')
    assert response.status == 200


@override_generic_settings(global_settings(), {'enabled': False})
def test_asgi_nr_disabled(application):
    response = application.get('/')
    assert response.status == 200



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest
import webtest
import copy

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task, BackgroundTask
from newrelic.api.transaction import (current_transaction, current_trace_id,
        current_span_id)
from newrelic.api.time_trace import current_trace
from newrelic.api.web_transaction import WSGIWebTransaction
from newrelic.api.wsgi_application import wsgi_application

from testing_support.fixtures import (override_application_settings,
        validate_attributes, 
        validate_error_event_attributes)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

distributed_trace_intrinsics = ['guid', 'traceId', 'priority', 'sampled']
inbound_payload_intrinsics = ['parent.type', 'parent.app', 'parent.account',
        'parent.transportType', 'parent.transportDuration']

payload = {
    'v': [0, 1],
    'd': {
        'ac': '1',
        'ap': '2827902',
        'id': '7d3efb1b173fecfa',
        'pa': '5e5733a911cfbc73',
        'pr': 10.001,
        'sa': True,
        'ti': 1518469636035,
        'tr': 'd6b4ba0c3a712ca',
        'ty': 'App',
    }
}
parent_order = ['parent_type', 'parent_account',
                'parent_app', 'parent_transport_type']
parent_info = {
    'parent_type': payload['d']['ty'],
    'parent_account': payload['d']['ac'],
    'parent_app': payload['d']['ap'],
    'parent_transport_type': 'HTTP'
}


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = '200 OK'
    output = b'hello world'
    response_headers = [('Content-type', 'text/html; charset=utf-8'),
                        ('Content-Length', str(len(output)))]

    txn = current_transaction()

    # Make assertions on the WSGIWebTransaction object
    assert txn._distributed_trace_state
    assert txn.parent_type == 'App'
    assert txn.parent_app == '2827902'
    assert txn.parent_account == '1'
    assert txn.parent_span == '7d3efb1b173fecfa'
    assert txn.parent_transport_type == 'HTTP'
    assert isinstance(txn.parent_transport_duration, float)
    assert txn._trace_id == 'd6b4ba0c3a712ca'
    assert txn.priority == 10.001
    assert txn.sampled

    start_response(status, response_headers)
    return [output]


test_application = webtest.TestApp(target_wsgi_application)

_override_settings = {
    'trusted_account_key': '1',
    'distributed_tracing.enabled': True,
}


_metrics = [
        ('Supportability/DistributedTrace/AcceptPayload/Success', 1),
        ('Supportability/TraceContext/Accept/Success', None)
]


@override_application_settings(_override_settings)
@validate_transaction_metrics(
    '',
    group='Uri',
    rollup_metrics=_metrics)
def test_distributed_tracing_web_transaction():
    headers = {'newrelic': json.dumps(payload)}
    response = test_application.get('/', headers=headers)
    assert 'X-NewRelic-App-Data' not in response.headers


@pytest.mark.parametrize('span_events', (True, False))
@pytest.mark.parametrize('accept_payload', (True, False))
def test_distributed_trace_attributes(span_events, accept_payload):
    if accept_payload:
        _required_intrinsics = (
                distributed_trace_intrinsics + inbound_payload_intrinsics)
        _forgone_txn_intrinsics = []
        _forgone_error_intrinsics = []
        _exact_intrinsics = {
            'parent.type': 'Mobile',
            'parent.app': '2827902',
            'parent.account': '1',
            'parent.transportType': 'HTTP',
            'traceId': 'd6b4ba0c3a712ca',
        }
        _exact_txn_attributes = {'agent': {}, 'user': {},
                'intrinsic': _exact_intrinsics.copy()}
        _exact_error_attributes = {'agent': {}, 'user': {},
                'intrinsic': _exact_intrinsics.copy()}
        _exact_txn_attributes['intrinsic']['parentId'] = '7d3efb1b173fecfa'
        _exact_txn_attributes['intrinsic']['parentSpanId'] = 'c86df80de2e6f51c'

        _forgone_error_intrinsics.append('parentId')
        _forgone_error_intrinsics.append('parentSpanId')
        _forgone_txn_intrinsics.append('grandparentId')
        _forgone_error_intrinsics.append('grandparentId')

        _required_attributes = {
                'intrinsic': _required_intrinsics, 'agent': [], 'user': []}
        _forgone_txn_attributes = {'intrinsic': _forgone_txn_intrinsics,
                'agent': [], 'user': []}
        _forgone_error_attributes = {'intrinsic': _forgone_error_intrinsics,
                'agent': [], 'user': []}
    else:
        _required_intrinsics = distributed_trace_intrinsics
        _forgone_txn_intrinsics = _forgone_error_intrinsics = \
                inbound_payload_intrinsics + ['grandparentId', 'parentId',
                'parentSpanId']

        _required_attributes = {
                'intrinsic': _required_intrinsics, 'agent': [], 'user': []}
        _forgone_txn_attributes = {'intrinsic': _forgone_txn_intrinsics,
                'agent': [], 'user': []}
        _forgone_error_attributes = {'intrinsic': _forgone_error_intrinsics,
                'agent': [], 'user': []}
        _exact_txn_attributes = _exact_error_attributes = None

    _forgone_trace_intrinsics = _forgone_error_intrinsics

    test_settings = _override_settings.copy()
    test_settings['span_events.enabled'] = span_events

    @override_application_settings(test_settings)
    @validate_transaction_event_attributes(
            _required_attributes, _forgone_txn_attributes,
            _exact_txn_attributes)
    @validate_error_event_attributes(
            _required_attributes, _forgone_error_attributes,
            _exact_error_attributes)
    @validate_attributes('intrinsic',
            _required_intrinsics, _forgone_trace_intrinsics)
    @background_task(name='test_distributed_trace_attributes')
    def _test():
        txn = current_transaction()

        payload = {
            "v": [0, 1],
            "d": {
                "ty": "Mobile",
                "ac": "1",
                "ap": "2827902",
                "id": "c86df80de2e6f51c",
                "tr": "d6b4ba0c3a712ca",
                "ti": 1518469636035,
                "tx": "7d3efb1b173fecfa"
            }
        }
        payload['d']['pa'] = "5e5733a911cfbc73"

        if accept_payload:
            result = txn.accept_distributed_trace_payload(payload)
            assert result
        else:
            txn._create_distributed_trace_payload()

        try:
            raise ValueError('cookies')
        except ValueError:
            txn.notice_error()

    _test()


_forgone_attributes = {
    'agent': [],
    'user': [],
    'intrinsic': (inbound_payload_intrinsics + ['grandparentId']),
}


@override_application_settings(_override_settings)
@validate_transaction_event_attributes(
        {}, _forgone_attributes)
@validate_error_event_attributes(
        {}, _forgone_attributes)
@validate_attributes('intrinsic',
        {}, _forgone_attributes['intrinsic'])
@background_task(name='test_distributed_trace_attrs_omitted')
def test_distributed_trace_attrs_omitted():
    txn = current_transaction()
    try:
        raise ValueError('cookies')
    except ValueError:
        txn.notice_error()


# test our distributed_trace metrics by creating a transaction and then forcing
# it to process a distributed trace payload
@pytest.mark.parametrize('web_transaction', (True, False))
@pytest.mark.parametrize('gen_error', (True, False))
@pytest.mark.parametrize('has_parent', (True, False))
def test_distributed_tracing_metrics(web_transaction, gen_error, has_parent):
    def _make_dt_tag(pi):
        return "%s/%s/%s/%s/all" % tuple(pi[x] for x in parent_order)

    # figure out which metrics we'll see based on the test params
    # note: we'll always see DurationByCaller if the distributed
    # tracing flag is turned on
    metrics = ['DurationByCaller']
    if gen_error:
        metrics.append('ErrorsByCaller')
    if has_parent:
        metrics.append('TransportDuration')

    tag = None
    dt_payload = copy.deepcopy(payload)

    # if has_parent is True, our metric name will be info about the parent,
    # otherwise it is Unknown/Unknown/Unknown/Unknown
    if has_parent:
        tag = _make_dt_tag(parent_info)
    else:
        tag = _make_dt_tag(dict((x, 'Unknown') for x in parent_info.keys()))
        del dt_payload['d']['tr']

    # now run the test
    transaction_name = "test_dt_metrics_%s" % '_'.join(metrics)
    _rollup_metrics = [
        ("%s/%s%s" % (x, tag, bt), 1)
        for x in metrics
        for bt in ['', 'Web' if web_transaction else 'Other']
    ]

    def _make_test_transaction():
        application = application_instance()

        if not web_transaction:
            return BackgroundTask(application, transaction_name)

        environ = {'REQUEST_URI': '/trace_ends_after_txn'}
        tn = WSGIWebTransaction(application, environ)
        tn.set_transaction_name(transaction_name)
        return tn

    @override_application_settings(_override_settings)
    @validate_transaction_metrics(
        transaction_name,
        background_task=not(web_transaction),
        rollup_metrics=_rollup_metrics)
    def _test():
        with _make_test_transaction() as transaction:
            transaction.accept_distributed_trace_payload(dt_payload)

            if gen_error:
                try:
                    1 / 0
                except ZeroDivisionError:
                    transaction.notice_error()

    _test()


NEW_RELIC_ACCEPTED = \
        [('Supportability/DistributedTrace/AcceptPayload/Success', 1),
         ('Supportability/TraceContext/Accept/Success', None),
         ('Supportability/TraceContext/TraceParent/Accept/Success', None),
         ('Supportability/TraceContext/Accept/Success', None)]
TRACE_CONTEXT_ACCEPTED = \
        [('Supportability/TraceContext/Accept/Success', 1),
         ('Supportability/TraceContext/TraceParent/Accept/Success', 1),
         ('Supportability/TraceContext/Accept/Success', 1),
         ('Supportability/DistributedTrace/AcceptPayload/Success', None)]
NO_HEADERS_ACCEPTED = \
        [('Supportability/DistributedTrace/AcceptPayload/Success', None),
        ('Supportability/TraceContext/Accept/Success', None),
        ('Supportability/TraceContext/TraceParent/Accept/Success', None),
        ('Supportability/TraceContext/Accept/Success', None)]
TRACEPARENT = '00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01'
TRACESTATE = 'rojo=f06a0ba902b7,congo=t61rcWkgMzE'


@pytest.mark.parametrize('traceparent,tracestate,newrelic,metrics',
                         [(False, False, False, NO_HEADERS_ACCEPTED),
                          (False, False, True, NEW_RELIC_ACCEPTED),
                          (False, True, True, NEW_RELIC_ACCEPTED),
                          (False, True, False, NO_HEADERS_ACCEPTED),
                          (True, True, True, TRACE_CONTEXT_ACCEPTED),
                          (True, False, False, TRACE_CONTEXT_ACCEPTED),
                          (True, False, True, TRACE_CONTEXT_ACCEPTED),
                          (True, True, False, TRACE_CONTEXT_ACCEPTED)]
                         )
@override_application_settings(_override_settings)
def test_distributed_tracing_backwards_compatibility(traceparent,
                                                     tracestate,
                                                     newrelic,
                                                     metrics):

    headers = []
    if traceparent:
        headers.append(('traceparent', TRACEPARENT))
    if tracestate:
        headers.append(('tracestate', TRACESTATE))
    if newrelic:
        headers.append(('newrelic', json.dumps(payload)))

    @validate_transaction_metrics(
        "test_distributed_tracing_backwards_compatibility",
        background_task=True,
        rollup_metrics=metrics)
    @background_task(name='test_distributed_tracing_backwards_compatibility')
    def _test():
        transaction = current_transaction()
        transaction.accept_distributed_trace_headers(headers)

    _test()


@background_task(name='test_current_trace_id_api_inside_transaction')
def test_current_trace_id_api_inside_transaction():
    trace_id = current_trace_id()
    assert len(trace_id) == 32
    assert trace_id == current_transaction().trace_id


def test_current_trace_id_api_outside_transaction():
    trace_id = current_trace_id()
    assert trace_id is None


@background_task(name='test_current_span_id_api_inside_transaction')
def test_current_span_id_inside_transaction():
    span_id = current_span_id()
    assert span_id == current_trace().guid


def test_current_span_id_outside_transaction():
    span_id = current_span_id()
    assert span_id is None



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import (
    core_application_stats_engine,
    override_application_settings,
    reset_core_stats_engine,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import BackgroundTask


@override_application_settings({"event_harvest_config.harvest_limits.analytic_event_data": 1})
@pytest.mark.parametrize("first_transaction_saved", [True, False])
def test_priority_used_in_transaction_events(first_transaction_saved):
    first_priority = 1 if first_transaction_saved else 0
    second_priority = 0 if first_transaction_saved else 1

    @reset_core_stats_engine()
    def _test():
        # Stats engine
        stats_engine = core_application_stats_engine()

        with BackgroundTask(application(), name="T1") as txn:
            txn._priority = first_priority

        with BackgroundTask(application(), name="T2") as txn:
            txn._priority = second_priority

        transaction_events = list(stats_engine.transaction_events)
        assert len(transaction_events) == 1

        # Highest priority should win.
        # Priority can be 1 or 2 depending on randomness in sampling computation.
        assert stats_engine.transaction_events.pq[0][0] >= 1

        if first_transaction_saved:
            assert transaction_events[0][0]["name"].endswith("/T1")
        else:
            assert transaction_events[0][0]["name"].endswith("/T2")

    _test()


@override_application_settings({"event_harvest_config.harvest_limits.error_event_data": 1})
@pytest.mark.parametrize("first_transaction_saved", [True, False])
def test_priority_used_in_transaction_error_events(first_transaction_saved):
    first_priority = 1 if first_transaction_saved else 0
    second_priority = 0 if first_transaction_saved else 1

    @reset_core_stats_engine()
    def _test():
        with BackgroundTask(application(), name="T1") as txn:
            txn._priority = first_priority
            try:
                raise ValueError("OOPS")
            except ValueError:
                txn.notice_error()

        with BackgroundTask(application(), name="T2") as txn:
            txn._priority = second_priority
            try:
                raise ValueError("OOPS")
            except ValueError:
                txn.notice_error()

        # Stats engine
        stats_engine = core_application_stats_engine()

        error_events = list(stats_engine.error_events)
        assert len(error_events) == 1

        # highest priority should win
        assert stats_engine.error_events.pq[0][0] == 1

        if first_transaction_saved:
            assert error_events[0][0]["transactionName"].endswith("/T1")
        else:
            assert error_events[0][0]["transactionName"].endswith("/T2")

    _test()


@override_application_settings({"event_harvest_config.harvest_limits.custom_event_data": 1})
@pytest.mark.parametrize("first_transaction_saved", [True, False])
def test_priority_used_in_transaction_custom_events(first_transaction_saved):
    first_priority = 1 if first_transaction_saved else 0
    second_priority = 0 if first_transaction_saved else 1

    @reset_core_stats_engine()
    def _test():
        with BackgroundTask(application(), name="T1") as txn:
            txn._priority = first_priority
            txn.record_custom_event("foobar", {"foo": "bar"})

        with BackgroundTask(application(), name="T2") as txn:
            txn._priority = second_priority
            txn.record_custom_event("barbaz", {"foo": "bar"})

        # Stats engine
        stats_engine = core_application_stats_engine()

        custom_events = list(stats_engine.custom_events)
        assert len(custom_events) == 1

        # highest priority should win
        assert stats_engine.custom_events.pq[0][0] == 1

        if first_transaction_saved:
            assert custom_events[0][0]["type"] == "foobar"
        else:
            assert custom_events[0][0]["type"] == "barbaz"

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import deque
from concurrent import futures
from threading import Event

import grpc

from newrelic.core.infinite_tracing_pb2 import RecordStatus, Span, SpanBatch

SPANS_PROCESSED_EVENT = Event()
SPANS_RECEIVED = deque()
SPAN_BATCHES_RECEIVED = deque()


def record_span(request, context):
    metadata = dict(context.invocation_metadata())
    assert "agent_run_token" in metadata
    assert "license_key" in metadata

    for span in request:
        SPANS_RECEIVED.append(span)
        SPANS_PROCESSED_EVENT.set()

        # Handle injecting status codes.
        status_code = span.intrinsics.get("status_code", None)
        status_code = status_code and getattr(grpc.StatusCode, status_code.string_value)
        if status_code is grpc.StatusCode.OK:
            return
        elif status_code:
            context.abort(status_code, "Abort triggered by client")

        # Give the client time to enter the wait condition before closing the server.
        if span.intrinsics.get("wait_then_ok", None):
            # Wait long enough that the client is now waiting for more spans and stuck in notify.wait().
            time.sleep(1)
            return

        yield RecordStatus(messages_seen=1)


def record_span_batch(request, context):
    metadata = dict(context.invocation_metadata())
    assert "agent_run_token" in metadata
    assert "license_key" in metadata

    for span_batch in request:
        SPAN_BATCHES_RECEIVED.append(span_batch)
        SPANS_PROCESSED_EVENT.set()
        batch_size = 0

        for span in span_batch.spans:
            # Handle injecting status codes.
            status_code = span.intrinsics.get("status_code", None)
            status_code = status_code and getattr(grpc.StatusCode, status_code.string_value)
            if status_code is grpc.StatusCode.OK:
                return
            elif status_code:
                context.abort(status_code, "Abort triggered by client")

            # Give the client time to enter the wait condition before closing the server.
            if span.intrinsics.get("wait_then_ok", None):
                # Wait long enough that the client is now waiting for more spans and stuck in notify.wait().
                time.sleep(1)
                return

        yield RecordStatus(messages_seen=batch_size)


HANDLERS = (
    grpc.method_handlers_generic_handler(
        "com.newrelic.trace.v1.IngestService",
        {
            "RecordSpan": grpc.stream_stream_rpc_method_handler(
                record_span, Span.FromString, RecordStatus.SerializeToString
            ),
            "RecordSpanBatch": grpc.stream_stream_rpc_method_handler(
                record_span_batch, SpanBatch.FromString, RecordStatus.SerializeToString
            ),
        },
    ),
)


def main():
    server = grpc.server(futures.ThreadPoolExecutor())
    server.add_generic_rpc_handlers(HANDLERS)
    server.add_insecure_port("127.0.0.1:8000")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    main()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import os.path
import ssl
import zlib

import pytest
from testing_support.mock_external_http_server import (
    BaseHTTPServer,
    MockExternalHTTPServer,
)

from newrelic.common import certs
from newrelic.common.agent_http import (
    ApplicationModeClient,
    DeveloperModeClient,
    HttpClient,
    InsecureHttpClient,
    ServerlessModeClient,
)
from newrelic.common.encoding_utils import ensure_str
from newrelic.common.object_names import callable_name
from newrelic.core.internal_metrics import InternalTraceContext
from newrelic.core.stats_engine import CustomMetrics
from newrelic.network.exceptions import NetworkInterfaceException
from newrelic.packages.urllib3.util import Url

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


SERVER_CERT = os.path.join(os.path.dirname(__file__), "cert.pem")


def echo_full_request(self):
    self.server.connections.append(self.connection)
    request_line = str(self.requestline).encode("utf-8")
    headers = "\n".join("%s: %s" % (k.lower(), v) for k, v in self.headers.items())
    self.send_response(200)
    self.end_headers()
    self.wfile.write(request_line)
    self.wfile.write(b"\n")
    self.wfile.write(headers.strip().encode("utf-8"))
    self.wfile.write(b"\n")
    content_length = int(self.headers.get("Content-Length", 0))
    if content_length:
        data = self.rfile.read(content_length)
        self.wfile.write(data)


class InsecureServer(MockExternalHTTPServer):
    def __init__(self, handler=echo_full_request, port=None, *args, **kwargs):
        super(MockExternalHTTPServer, self).__init__(*args, **kwargs)
        self.port = port or self.get_open_port()

        def do_CONNECT(self):
            host_port = self.requestline.split(" ", 2)[1]
            self.server.connect_host, self.server.connect_port = host_port.split(":", 1)
            self.server.connect_headers = self.headers

            self.send_response(200)
            self.end_headers()
            self.close_connection = True

        handler = type(
            "ResponseHandler",
            (
                BaseHTTPServer.BaseHTTPRequestHandler,
                object,
            ),
            {"do_GET": handler, "do_POST": handler, "do_CONNECT": do_CONNECT},
        )
        self.httpd = BaseHTTPServer.HTTPServer(("localhost", self.port), handler)
        self.httpd.connections = []
        self.httpd.connect_host = None
        self.httpd.connect_port = None
        self.httpd.connect_headers = {}
        self.daemon = True

    def reset(self):
        self.httpd.connect_host = None
        self.httpd.connect_port = None
        self.httpd.connect_headers = {}


class SecureServer(InsecureServer):
    def __init__(self, *args, **kwargs):
        super(SecureServer, self).__init__(*args, **kwargs)
        try:
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            self.context.load_cert_chain(certfile=SERVER_CERT, keyfile=SERVER_CERT)
            self.httpd.socket = self.context.wrap_socket(
                sock=self.httpd.socket,
                server_side=True,
                do_handshake_on_connect=False,
            )
        except (AttributeError, TypeError):
            self.httpd.socket = ssl.wrap_socket(
                self.httpd.socket,
                server_side=True,
                keyfile=SERVER_CERT,
                certfile=SERVER_CERT,
                do_handshake_on_connect=False,
            )


@pytest.fixture(scope="module")
def server():
    with SecureServer() as server:
        yield server


@pytest.fixture(scope="module")
def insecure_server():
    with InsecureServer() as server:
        yield server


@pytest.mark.parametrize(
    "scheme,host,port,username,password,expected",
    (
        (None, "host", 0, None, None, None),
        ("http", None, 8080, None, None, None),
        ("http", "host", 0, None, None, Url(scheme="http", host="host", port=None)),
        (
            "http",
            "host",
            8080,
            None,
            None,
            Url(scheme="http", host="host", port=8080),
        ),
        (
            "http",
            "https://host:8081",
            8080,
            None,
            None,
            Url(scheme="https", host="host", port=8081),
        ),
        (
            "http",
            "https://user:pass@host:8081",
            8080,
            None,
            None,
            Url(scheme="https", host="host", port=8081, auth="user:pass"),
        ),
        (
            "https",
            "host",
            8081,
            "username",
            None,
            Url(scheme="https", auth="username", host="host", port=8081),
        ),
        (
            "https",
            "host",
            8081,
            "username",
            "password",
            Url(scheme="https", auth="username:password", host="host", port=8081),
        ),
    ),
)
def test_proxy_parsing(scheme, host, port, username, password, expected):
    assert HttpClient._parse_proxy(scheme, host, port, username, password) == expected


@pytest.mark.parametrize("method", ("GET", "POST"))
def test_http_no_payload(server, method):
    with HttpClient("localhost", server.port, disable_certificate_validation=True) as client:
        connection = client._connection_attr
        status, data = client.send_request(method=method, headers={"foo": "bar"})

    assert status == 200
    data = ensure_str(data)
    data = data.split("\n")

    # Verify connection has been closed
    assert client._connection_attr is None
    assert connection.pool is None

    # Verify request line
    assert data[0].startswith(method + " /agent_listener/invoke_raw_method ")

    # Verify headers
    user_agent_header = ""
    foo_header = ""

    for header in data[1:-1]:
        if header.lower().startswith("user-agent:"):
            _, value = header.split(":", 1)
            value = value.strip()
            user_agent_header = value
        elif header.startswith("foo:"):
            _, value = header.split(":", 1)
            value = value.strip()
            foo_header = value

    assert user_agent_header.startswith("NewRelic-PythonAgent/")
    assert foo_header == "bar"


@pytest.mark.parametrize("client_cls", (HttpClient, ApplicationModeClient))
def test_non_ok_response(client_cls, server):
    internal_metrics = CustomMetrics()

    with client_cls("localhost", server.port, disable_certificate_validation=True) as client:
        with InternalTraceContext(internal_metrics):
            status, _ = client.send_request(method="PUT")

    assert status != 200
    internal_metrics = dict(internal_metrics.metrics())

    if client_cls is ApplicationModeClient:
        assert internal_metrics == {
            "Supportability/Python/Collector/Failures": [1, 0, 0, 0, 0, 0],
            "Supportability/Python/Collector/Failures/direct": [1, 0, 0, 0, 0, 0],
            "Supportability/Python/Collector/HTTPError/%d" % status: [1, 0, 0, 0, 0, 0],
        }
    else:
        assert not internal_metrics


def test_http_close_connection(server):
    client = HttpClient(
        "localhost",
        server.port,
        disable_certificate_validation=True,
    )

    status, _ = client.send_request()
    assert status == 200

    connection = client._connection_attr
    client.close_connection()
    assert client._connection_attr is None
    assert connection.pool is None

    # Verify that subsequent calls to close_connection don't crash
    client.close_connection()


def test_http_close_connection_in_context_manager():
    client = HttpClient("localhost", 1000)
    with client:
        client.close_connection()


@pytest.mark.parametrize(
    "client_cls,method,threshold",
    (
        (HttpClient, "gzip", 0),
        (HttpClient, "gzip", 100),
        (ApplicationModeClient, "gzip", 0),
        (ApplicationModeClient, "gzip", 100),
        (ApplicationModeClient, "deflate", 0),
        (ApplicationModeClient, "deflate", 100),
    ),
)
def test_http_payload_compression(server, client_cls, method, threshold):
    payload = b"*" * 20

    internal_metrics = CustomMetrics()

    with client_cls(
        "localhost",
        server.port,
        disable_certificate_validation=True,
        compression_method=method,
        compression_threshold=threshold,
    ) as client:
        with InternalTraceContext(internal_metrics):
            status, data = client.send_request(payload=payload, params={"method": "method1"})

    # Sending one additional request to valid metric aggregation for top level data usage supportability metrics
    with client_cls(
        "localhost",
        server.port,
        disable_certificate_validation=True,
        compression_method=method,
        compression_threshold=threshold,
    ) as client:
        with InternalTraceContext(internal_metrics):
            status, data = client.send_request(payload=payload, params={"method": "method2"})

    assert status == 200
    data = data.split(b"\n")
    sent_payload = data[-1]
    payload_byte_len = len(sent_payload)
    internal_metrics = dict(internal_metrics.metrics())
    if client_cls is ApplicationModeClient:
        assert internal_metrics["Supportability/Python/Collector/method1/Output/Bytes"][:2] == [
            1,
            len(payload),
        ]
        assert internal_metrics["Supportability/Python/Collector/Output/Bytes"][:2] == [
            2,
            len(payload)*2,
        ]

        if threshold < 20:
            # Verify compression time is recorded
            assert internal_metrics["Supportability/Python/Collector/method1/ZLIB/Compress"][0] == 1
            assert internal_metrics["Supportability/Python/Collector/method1/ZLIB/Compress"][1] > 0

            # Verify the compressed payload length is recorded
            assert internal_metrics["Supportability/Python/Collector/method1/ZLIB/Bytes"][:2] == [1, payload_byte_len]
            assert internal_metrics["Supportability/Python/Collector/ZLIB/Bytes"][:2] == [2, payload_byte_len*2]
            
            assert len(internal_metrics) == 8
        else:
            # Verify no ZLIB compression metrics were sent
            assert len(internal_metrics) == 3
    else:
        assert not internal_metrics

    if threshold < 20:
        expected_content_encoding = method.encode("utf-8")
        assert sent_payload != payload
        if method == "deflate":
            sent_payload = zlib.decompress(sent_payload)
        elif method == "gzip":
            decompressor = zlib.decompressobj(31)
            sent_payload = decompressor.decompress(sent_payload)
            sent_payload += decompressor.flush()
    else:
        expected_content_encoding = b"Identity"

    for header in data[1:-1]:
        if header.lower().startswith(b"content-encoding"):
            _, content_encoding = header.split(b":", 1)
            content_encoding = content_encoding.strip()
            break
    else:
        assert False, "Missing content-encoding header"

    assert content_encoding == expected_content_encoding
    assert sent_payload == payload


def test_cert_path(server):
    with HttpClient("localhost", server.port, ca_bundle_path=SERVER_CERT) as client:
        status, data = client.send_request()


@pytest.mark.parametrize("system_certs_available", (True, False))
def test_default_cert_path(monkeypatch, system_certs_available):
    if system_certs_available:
        cert_file = "foo"
    else:
        cert_file = None

    class DefaultVerifyPaths(object):
        cafile = cert_file

        def __init__(self, *args, **kwargs):
            pass

    monkeypatch.setattr(ssl, "DefaultVerifyPaths", DefaultVerifyPaths)
    internal_metrics = CustomMetrics()
    with InternalTraceContext(internal_metrics):
        client = HttpClient("localhost", ca_bundle_path=None)

    internal_metrics = dict(internal_metrics.metrics())
    cert_metric = "Supportability/Python/Certificate/BundleRequired"
    if system_certs_available:
        assert "ca_certs" not in client._connection_kwargs
        assert cert_metric not in internal_metrics
    else:
        assert client._connection_kwargs["ca_certs"] == certs.where()
        assert internal_metrics[cert_metric][-3:-1] == [1, 1]


@pytest.mark.parametrize("auth", ((None, None), ("username", None), ("username", "password")))
def test_ssl_via_ssl_proxy(server, auth):
    proxy_user, proxy_pass = auth
    with HttpClient(
        "localhost",
        1,
        proxy_scheme="https",
        proxy_host="localhost",
        proxy_port=server.port,
        proxy_user=proxy_user,
        proxy_pass=proxy_pass,
        disable_certificate_validation=True,
    ) as client:
        status, data = client.send_request()

    assert status == 200
    data = data.decode("utf-8")
    data = data.split("\n")
    assert data[0].startswith("POST https://localhost:1/agent_listener/invoke_raw_method ")

    proxy_auth = None
    for header in data[1:-1]:
        if header.lower().startswith("proxy-authorization"):
            _, proxy_auth = header.split(":", 1)
            proxy_auth = proxy_auth.strip()
            break

    if proxy_user:
        auth_expected = proxy_user
        if proxy_pass:
            auth_expected = auth_expected + ":" + proxy_pass
        auth_expected = "Basic " + base64.b64encode(auth_expected.encode("utf-8")).decode("utf-8")
        assert proxy_auth == auth_expected
    else:
        assert not proxy_auth

    assert server.httpd.connect_host is None


def test_non_ssl_via_ssl_proxy(server):
    with InsecureHttpClient(
        "localhost",
        1,
        proxy_scheme="https",
        proxy_host="localhost",
        proxy_port=server.port,
        disable_certificate_validation=True,
    ) as client:
        status, data = client.send_request()

    assert status == 200
    data = data.decode("utf-8")
    data = data.split("\n")
    assert data[0].startswith("POST http://localhost:1/agent_listener/invoke_raw_method ")

    assert server.httpd.connect_host is None


def test_non_ssl_via_non_ssl_proxy(insecure_server):
    with InsecureHttpClient(
        "localhost",
        1,
        proxy_scheme="http",
        proxy_host="localhost",
        proxy_port=insecure_server.port,
    ) as client:
        status, data = client.send_request()

    assert status == 200
    data = data.decode("utf-8")
    data = data.split("\n")
    assert data[0].startswith("POST http://localhost:1/agent_listener/invoke_raw_method ")

    assert insecure_server.httpd.connect_host is None


@pytest.mark.parametrize("auth", ((None, None), ("username", None), ("username", "password")))
def test_ssl_via_non_ssl_proxy(insecure_server, auth):
    proxy_user, proxy_pass = auth
    with HttpClient(
        "localhost",
        1,
        proxy_scheme="http",
        proxy_host="localhost",
        proxy_port=insecure_server.port,
        proxy_user=proxy_user,
        proxy_pass=proxy_pass,
        disable_certificate_validation=True,
    ) as client:
        try:
            client.send_request()
        except Exception:
            pass

    try:
        if proxy_user:
            auth_expected = proxy_user
            if proxy_pass:
                auth_expected = auth_expected + ":" + proxy_pass
            auth_expected = "Basic " + base64.b64encode(auth_expected.encode("utf-8")).decode("utf-8")
            assert insecure_server.httpd.connect_headers["proxy-authorization"] == auth_expected
        else:
            assert "proxy-authorization" not in insecure_server.httpd.connect_headers
        assert insecure_server.httpd.connect_host == "localhost"
        assert insecure_server.httpd.connect_port == "1"
    finally:
        insecure_server.reset()


def test_max_payload_does_not_send(insecure_server):
    with InsecureHttpClient("localhost", insecure_server.port, max_payload_size_in_bytes=0) as client:
        status, data = client.send_request(payload=b"*")

    assert status == 413
    assert not data


@pytest.mark.parametrize(
    "method",
    (
        "preconnect",
        "connect",
        "agent_settings",
        "get_agent_commands",
        "agent_command_results",
        "metric_data",
        "analytic_event_data",
        "span_event_data",
        "error_event_data",
        "custom_event_data",
        "error_data",
        "transaction_sample_data",
        "sql_trace_data",
        "profile_data",
        "shutdown",
        "INVALID",
        None,
    ),
)
def test_developer_mode_client(method):
    params = {}
    with DeveloperModeClient("localhost", 1) as client:
        if method:
            params["method"] = method
        status, data = client.send_request(params=params)

    if not method or method == "INVALID":
        assert status == 400
    else:
        assert status == 200
        assert json.loads(data.decode("utf-8"))


def test_serverless_mode_client():
    methods = (
        "metric_data",
        "analytic_event_data",
        "span_event_data",
        "error_event_data",
        "custom_event_data",
        "error_data",
        "transaction_sample_data",
        "sql_trace_data",
    )
    with ServerlessModeClient("localhost", 1) as client:
        for method in methods:
            params = {"method": method}
            status, data = client.send_request(
                params=params,
                payload=json.dumps({"method": method}).encode("utf-8"),
            )

            assert status == 200
            assert json.loads(data.decode("utf-8"))

        # Verify that missing methods aren't captured
        status, _ = client.send_request(payload=b"{}")
        assert status == 400

        # Verify that invalid methods aren't captured
        status, _ = client.send_request(params={"method": "foo"}, payload=b"{}")
        assert status == 400

    payloads = client.finalize()
    assert len(payloads) == len(methods)
    for method in methods:
        assert payloads[method] == {"method": method}


@pytest.mark.parametrize(
    "client_cls,proxy_host,exception",
    (
        (HttpClient, "localhost", False),
        (HttpClient, None, False),
        (HttpClient, None, True),
        (ApplicationModeClient, None, True),
        (ApplicationModeClient, "localhost", True),
        (DeveloperModeClient, None, False),
        (ServerlessModeClient, None, False),
        (InsecureHttpClient, None, False),
    ),
)
def test_audit_logging(server, insecure_server, client_cls, proxy_host, exception):
    audit_log_fp = StringIO()
    params = {"method": "metric_data"}
    prefix = getattr(client_cls, "PREFIX_SCHEME", "https://")
    if exception:
        port = MockExternalHTTPServer.get_open_port()
    elif prefix == "https://":
        port = server.port
    else:
        port = insecure_server.port

    internal_metrics = CustomMetrics()

    with client_cls(
        "localhost",
        port,
        proxy_scheme="https",
        proxy_host=proxy_host,
        proxy_port=server.port if not exception else port,
        audit_log_fp=audit_log_fp,
        disable_certificate_validation=True,
    ) as client:
        with InternalTraceContext(internal_metrics):
            try:
                client.send_request(params=params)
                exc = ""
            except Exception as e:
                exc = callable_name(type(e.args[0]))

    internal_metrics = dict(internal_metrics.metrics())
    if exception and client_cls is ApplicationModeClient:
        if proxy_host:
            connection = "https-proxy"
        else:
            connection = "direct"
        assert internal_metrics == {
            "Supportability/Python/Collector/Failures": [1, 0, 0, 0, 0, 0],
            "Supportability/Python/Collector/Failures/%s" % connection: [1, 0, 0, 0, 0, 0],
            "Supportability/Python/Collector/Exception/%s" % exc: [1, 0, 0, 0, 0, 0],
        }
    else:
        assert not internal_metrics

    # Verify the audit log isn't empty
    assert audit_log_fp.tell()

    audit_log_fp.seek(0)
    audit_log_contents = audit_log_fp.read()
    assert prefix in audit_log_contents


def test_closed_connection():
    with HttpClient("localhost", MockExternalHTTPServer.get_open_port()) as client:
        with pytest.raises(NetworkInterfaceException):
            client.send_request()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pwd

USER = pwd.getpwuid(os.getuid()).pw_name


def postgresql_settings():
    """Return a list of dict of settings for connecting to postgresql.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2

        user = password = db = "postgres"
        base_port = 8080
    else:
        instances = 1

        user = db = USER
        password = ""
        base_port = 5432

    settings = [
        {
            "user": user,
            "password": password,
            "name": db,
            "host": "localhost",
            "port": base_port + instance_num,
            "table_name": "postgres_table_" + str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def mysql_settings():
    """Return a list of dict of settings for connecting to postgresql.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2

        user = password = db = "python_agent"
        base_port = 8080
    else:
        instances = 1

        user = db = USER
        password = ""
        base_port = 3306

    settings = [
        {
            "user": user,
            "password": password,
            "name": db,
            "host": "127.0.0.1",
            "port": base_port + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def redis_settings():
    """Return a list of dict of settings for connecting to redis.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 6379

    settings = [
        {
            "host": "localhost",
            "port": base_port + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def memcached_settings():
    """Return a list of dict of settings for connecting to memcached.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 11211

    settings = [
        {
            "host": "127.0.0.1",
            "port": base_port + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def mongodb_settings():
    """Return a list of dict of settings for connecting to mongodb.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 27017

    settings = [
        {"host": "127.0.0.1", "port": base_port + instance_num, "collection": "mongodb_collection_" + str(os.getpid())}
        for instance_num in range(instances)
    ]
    return settings


def elasticsearch_settings():
    """Return a list of dict of settings for connecting to elasticsearch.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 9200

    settings = [
        {
            "host": "localhost",
            "port": str(base_port + instance_num),
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def solr_settings():
    """Return a list of dict of settings for connecting to solr.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 8983

    settings = [
        {
            "host": "127.0.0.1",
            "port": base_port + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def rabbitmq_settings():
    """Return a list of dict of settings for connecting to rabbitmq.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    instances = 1
    base_port = 5672

    settings = [
        {
            "host": "localhost",
            "port": base_port + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def kafka_settings():
    """Return a list of dict of settings for connecting to kafka.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    if "GITHUB_ACTIONS" in os.environ:
        instances = 2
        base_port = 8080
    else:
        instances = 1
        base_port = 9092

    settings = [
        {
            "host": "localhost",
            "port": base_port + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.lambda_handler import lambda_handler
from newrelic.api.transaction import current_transaction
from newrelic.core.config import global_settings

from testing_support.fixtures import override_generic_settings
from testing_support.validators.validate_serverless_data import (
        validate_serverless_data)
from testing_support.validators.validate_serverless_payload import (
        validate_serverless_payload)
from testing_support.validators.validate_serverless_metadata import (
        validate_serverless_metadata)


@pytest.fixture(scope='function')
def serverless_application(request):
    settings = global_settings()
    orig = settings.serverless_mode.enabled
    settings.serverless_mode.enabled = True

    application_name = 'Python Agent Test (test_serverless_mode:%s)' % (
            request.node.name)
    application = application_instance(application_name)
    application.activate()

    yield application

    settings.serverless_mode.enabled = orig


def test_serverless_payload(capsys, serverless_application):

    @override_generic_settings(serverless_application.settings, {
        'distributed_tracing.enabled': True,
    })
    @validate_serverless_data(
            expected_methods=('metric_data', 'analytic_event_data'),
            forgone_methods=('preconnect', 'connect', 'get_agent_commands'))
    @validate_serverless_payload()
    @background_task(
            application=serverless_application,
            name='test_serverless_payload')
    def _test():
        transaction = current_transaction()
        assert transaction.settings.serverless_mode.enabled

    _test()

    out, err = capsys.readouterr()

    # Validate that something is printed to stdout
    assert out

    # Verify that the payload is loadable JSON
    payload = json.loads(out)


def test_no_cat_headers(serverless_application):
    @background_task(
            application=serverless_application,
            name='test_cat_headers')
    def _test_cat_headers():
        transaction = current_transaction()

        payload = ExternalTrace.generate_request_headers(transaction)
        assert not payload

        trace = ExternalTrace('testlib', 'http://example.com')
        response_headers = [('X-NewRelic-App-Data', 'Cookies')]
        with trace:
            trace.process_response_headers(response_headers)

        assert transaction.settings.cross_application_tracer.enabled is False

    _test_cat_headers()


def test_dt_outbound(serverless_application):
    @override_generic_settings(serverless_application.settings, {
        'distributed_tracing.enabled': True,
        'account_id': '1',
        'trusted_account_key': '1',
        'primary_application_id': '1',
    })
    @background_task(
            application=serverless_application,
            name='test_dt_outbound')
    def _test_dt_outbound():
        transaction = current_transaction()
        payload = ExternalTrace.generate_request_headers(transaction)
        assert payload

    _test_dt_outbound()


def test_dt_inbound(serverless_application):
    @override_generic_settings(serverless_application.settings, {
        'distributed_tracing.enabled': True,
        'account_id': '1',
        'trusted_account_key': '1',
        'primary_application_id': '1',
    })
    @background_task(
            application=serverless_application,
            name='test_dt_inbound')
    def _test_dt_inbound():
        transaction = current_transaction()

        payload = {
            'v': [0, 1],
            'd': {
                'ty': 'Mobile',
                'ac': '1',
                'tk': '1',
                'ap': '2827902',
                'pa': '5e5733a911cfbc73',
                'id': '7d3efb1b173fecfa',
                'tr': 'd6b4ba0c3a712ca',
                'ti': 1518469636035,
                'tx': '8703ff3d88eefe9d',
            }
        }

        result = transaction.accept_distributed_trace_payload(payload)
        assert result

    _test_dt_inbound()


@pytest.mark.parametrize('arn_set', (True, False))
def test_payload_metadata_arn(serverless_application, arn_set):

    # If the session object gathers the arn from the settings object before the
    # lambda handler records it there, then this test will fail.

    settings = global_settings()
    original_metadata = settings.aws_lambda_metadata.copy()

    arn = None
    if arn_set:
        arn = 'arrrrrrrrrrRrrrrrrrn'

    settings.aws_lambda_metadata.update({'arn': arn, 'function_version': '$LATEST'})

    class Context(object):
        invoked_function_arn = arn

    @validate_serverless_metadata(exact_metadata={'arn': arn})
    @lambda_handler(application=serverless_application)
    def handler(event, context):
        assert settings.aws_lambda_metadata['arn'] == arn
        return {}

    try:
        handler({}, Context)
    finally:
        settings.aws_lambda_metadata = original_metadata



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

try:
    from urllib2 import urlopen  # Py2.X
except ImportError:
    from urllib.request import urlopen  # Py3.X

import sqlite3 as db

from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import (
    add_custom_attribute,
    get_browser_timing_footer,
    get_browser_timing_header,
    record_custom_event,
)
from newrelic.api.wsgi_application import wsgi_application

_logger = logging.getLogger(__name__)

_custom_parameters = {
    "user": "user-name",
    "account": "account-name",
    "product": "product-name",
    "bytes": b"bytes-value",
    "string": "string-value",
    "unicode": "unicode-value",
    "integer": 1,
    "float": 1.0,
    "invalid-utf8": b"\xe2",
    "multibyte-utf8": b"\xe2\x88\x9a",
    "multibyte-unicode": b"\xe2\x88\x9a".decode("utf-8"),
    "list": [],
    "tuple": (),
    "dict": {},
}

_err_param = {"err-param": "value"}


def user_attributes_added():
    """Expected values when the custom parameters in this file are added as user
    attributes
    """
    user_attributes = _custom_parameters.copy()
    user_attributes["list"] = "[]"
    user_attributes["tuple"] = "()"
    user_attributes["dict"] = "{}"
    return user_attributes


def error_user_params_added():
    return _err_param.copy()


@wsgi_application()
def fully_featured_app(environ, start_response):
    status = "200 OK"

    path = environ.get("PATH_INFO")
    use_user_attrs = environ.get("record_attributes", "TRUE") == "TRUE"

    environ["wsgi.input"].read()
    environ["wsgi.input"].readline()
    environ["wsgi.input"].readlines()

    if use_user_attrs:

        for attr, val in _custom_parameters.items():
            add_custom_attribute(attr, val)

    if "db" in environ and int(environ["db"]) > 0:
        connection = db.connect(":memory:")
        for i in range(int(environ["db"]) - 1):
            connection.execute("create table test_db%d (a, b, c)" % i)

    if "external" in environ:
        for i in range(int(environ["external"])):
            r = urlopen("http://www.python.org")  # nosec
            r.read(10)

    if "err_message" in environ:
        n_errors = int(environ.get("n_errors", 1))
        for i in range(n_errors):
            try:

                # append number to stats engine to get unique errors, so they
                # don't immediately get filtered out.

                raise ValueError(environ["err_message"] + str(i))
            except ValueError:
                if use_user_attrs:
                    notice_error(attributes=_err_param)
                else:
                    notice_error()

    text = "<html><head>%s</head><body><p>RESPONSE</p>%s</body></html>"

    output = (text % (get_browser_timing_header(), get_browser_timing_footer())).encode("UTF-8")

    response_headers = [("Content-type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    write = start_response(status, response_headers)

    write(b"")

    return [output]


@wsgi_application()
def simple_exceptional_app(environ, start_response):

    start_response("500 :(", [])

    raise ValueError("Transaction had bad value")


def simple_app_raw(environ, start_response):
    status = "200 OK"

    _logger.info("Starting response")
    start_response(status, response_headers=[])

    return []


simple_app = wsgi_application()(simple_app_raw)


@wsgi_application()
def simple_custom_event_app(environ, start_response):

    params = {"snowman": "\u2603", "foo": "bar"}
    record_custom_event("SimpleAppEvent", params)

    start_response(status="200 OK", response_headers=[])
    return []



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction, record_log_event, ignore_transaction
from testing_support.fixtures import override_application_settings, reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import validate_log_event_count_outside_transaction
from testing_support.validators.validate_log_events import validate_log_events
from testing_support.validators.validate_log_events_outside_transaction import validate_log_events_outside_transaction


def set_trace_ids():
    txn = current_transaction()
    if txn:
        txn._trace_id = "abcdefgh12345678"
    trace = current_trace()
    if trace:
        trace.guid = "abcdefgh"


def exercise_record_log_event(message="A"):
    set_trace_ids()

    record_log_event(message, "ERROR")

enable_log_forwarding = override_application_settings({"application_logging.forwarding.enabled": True})
disable_log_forwarding = override_application_settings({"application_logging.forwarding.enabled": False})

_common_attributes_service_linking = {"timestamp": None, "hostname": None, "entity.name": "Python Agent Test (agent_features)", "entity.guid": None}
_common_attributes_trace_linking = {"span.id": "abcdefgh", "trace.id": "abcdefgh12345678"}
_common_attributes_trace_linking.update(_common_attributes_service_linking)
_test_record_log_event_inside_transaction_events = [{"message": "A", "level": "ERROR"}]
_test_record_log_event_inside_transaction_events[0].update(_common_attributes_trace_linking)

@enable_log_forwarding
def test_record_log_event_inside_transaction():
    @validate_log_events(_test_record_log_event_inside_transaction_events)
    @validate_log_event_count(1)
    @background_task()
    def test():
        exercise_record_log_event()
    
    test()


_test_record_log_event_outside_transaction_events = [{"message": "A", "level": "ERROR"}]
_test_record_log_event_outside_transaction_events[0].update(_common_attributes_service_linking)

@enable_log_forwarding
@reset_core_stats_engine()
def test_record_log_event_outside_transaction():
    @validate_log_events_outside_transaction(_test_record_log_event_outside_transaction_events)
    @validate_log_event_count_outside_transaction(1)
    def test():
        exercise_record_log_event()

    test()


_test_record_log_event_unknown_level_inside_transaction_events = [{"message": "A", "level": "UNKNOWN"}]
_test_record_log_event_unknown_level_inside_transaction_events[0].update(_common_attributes_trace_linking)

@enable_log_forwarding
def test_record_log_event_unknown_level_inside_transaction():
    @validate_log_events(_test_record_log_event_unknown_level_inside_transaction_events)
    @validate_log_event_count(1)
    @background_task()
    def test():
        set_trace_ids()
        record_log_event("A")
    
    test()


_test_record_log_event_unknown_level_outside_transaction_events = [{"message": "A", "level": "UNKNOWN"}]
_test_record_log_event_unknown_level_outside_transaction_events[0].update(_common_attributes_service_linking)

@enable_log_forwarding
@reset_core_stats_engine()
def test_record_log_event_unknown_level_outside_transaction():
    @validate_log_events_outside_transaction(_test_record_log_event_unknown_level_outside_transaction_events)
    @validate_log_event_count_outside_transaction(1)
    def test():
        set_trace_ids()
        record_log_event("A")

    test()


@enable_log_forwarding
def test_record_log_event_empty_message_inside_transaction():
    @validate_log_event_count(0)
    @background_task()
    def test():
        exercise_record_log_event("")
    
    test()


@enable_log_forwarding
@reset_core_stats_engine()
def test_record_log_event_empty_message_outside_transaction():
    @validate_log_event_count_outside_transaction(0)
    def test():
        exercise_record_log_event("")

    test()


@enable_log_forwarding
def test_record_log_event_whitespace_inside_transaction():
    @validate_log_event_count(0)
    @background_task()
    def test():
        exercise_record_log_event("         ")

    test()


@enable_log_forwarding
@reset_core_stats_engine()
def test_record_log_event_whitespace_outside_transaction():
    @validate_log_event_count_outside_transaction(0)
    def test():
        exercise_record_log_event("         ")

    test()


@enable_log_forwarding
def test_ignored_transaction_logs_not_forwarded():
    @validate_log_event_count(0)
    @background_task()
    def test():
        ignore_transaction()
        exercise_record_log_event()

    test()


_test_log_event_truncation_events = [{"message": "A" * 32768, "level": "ERROR"}]
_test_log_event_truncation_events[0].update(_common_attributes_trace_linking)

@enable_log_forwarding
def test_log_event_truncation():
    @validate_log_events(_test_log_event_truncation_events)
    @validate_log_event_count(1)
    @background_task()
    def test():
        exercise_record_log_event("A" * 33000)

    test()


@disable_log_forwarding
def test_record_log_event_inside_transaction():
    @validate_log_event_count(0)
    @background_task()
    def test():
        exercise_record_log_event()
    
    test()


@disable_log_forwarding
@reset_core_stats_engine()
def test_record_log_event_outside_transaction():
    @validate_log_event_count_outside_transaction(0)
    def test():
        exercise_record_log_event()

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest
from newrelic.api.application import application_instance
from newrelic.api.transaction import Transaction, current_transaction
from newrelic.api.background_task import background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.common.encoding_utils import (
        DistributedTracePayload, W3CTraceParent, W3CTraceState, NrTraceState)

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import (
        validate_span_events)
from _test_common import create_request, wait_for_transaction_completion
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

_test_matrix = ('method_name,streaming_request', (
    ('DoUnaryUnary', False),
    ('DoUnaryStream', False),
    ('DoStreamUnary', True),
    ('DoStreamStream', True)
))


@override_application_settings({'distributed_tracing.enabled': True})
@pytest.mark.parametrize(*_test_matrix)
def test_inbound_distributed_trace(mock_grpc_server, method_name,
        streaming_request, stub):
    request = create_request(streaming_request)

    transaction = Transaction(application_instance())
    dt_headers = ExternalTrace.generate_request_headers(transaction)

    @validate_transaction_metrics(
        'sample_application:SampleApplicationServicer.' + method_name,
        rollup_metrics=(
            ('Supportability/TraceContext/Accept/Success', 1),
        ),
    )
    @wait_for_transaction_completion
    def _test():
        method = getattr(stub, method_name)
        response = method(request, metadata=dt_headers)

        try:
            list(response)
        except Exception:
            pass

    _test()


_test_matrix = ('method_type,method_name', (
        ('unary_unary', '__call__'),
        ('unary_unary', 'with_call'),
        ('stream_unary', '__call__'),
        ('stream_unary', 'with_call'),
        ('unary_stream', '__call__'),
        ('stream_stream', '__call__'),
))


@pytest.mark.parametrize(*_test_matrix)
@pytest.mark.parametrize('dt_enabled,dt_error', (
    (True, False),
    (True, True),
    (False, False),
))
def test_outbound_distributed_trace(
        mock_grpc_server, method_type, method_name, dt_enabled, dt_error, stub):
    request_type, response_type = method_type.split('_', 1)
    streaming_request = request_type == 'stream'
    streaming_response = response_type == 'stream'
    stub_method = 'DtNoTxn' + method_type.title().replace('_', '')

    request = create_request(streaming_request)

    method_callable = getattr(stub, stub_method)
    method = getattr(method_callable, method_name)

    exact_intrinsics = {
        'category': 'http',
        'span.kind': 'client',
    }

    txn_name = 'test_outbound_DT[{0}-{1}-{2}-{3}]'.format(
            method_type, method_name, dt_enabled, dt_error)
    settings = {'distributed_tracing.enabled': dt_enabled}
    span_count = 1 if dt_enabled else 0
    if dt_error:
        settings['trusted_account_key'] = None

    @override_application_settings(settings)
    @validate_span_events(count=span_count, exact_intrinsics=exact_intrinsics)
    @wait_for_transaction_completion
    @background_task(name=txn_name)
    def _test():
        # Always mark sampled as True. current_transaction() is guaranteed to
        # be non-None here as test fixtures ensure activation prior to testing.
        current_transaction()._sampled = True

        reply = method(request)

        if isinstance(reply, tuple):
            reply = reply[0]

        try:
            # If the reply was canceled or the server code raises an exception,
            # this will raise an exception which will be recorded by the agent
            if streaming_response:
                reply = list(reply)
            else:
                reply = [reply.result()]
        except (AttributeError, TypeError):
            reply = [reply]

        metadata = json.loads(reply[0].text)

        if not dt_enabled or dt_error:
            assert 'newrelic' not in metadata
            assert 'traceparent' not in metadata
            assert 'tracestate' not in metadata
        else:
            decoded = DistributedTracePayload.decode(metadata['newrelic'])

            # The external span should be the parent
            exact_intrinsics['guid'] = decoded['d']['id']

            # Check that tracestate / traceparent payload matches newrelic
            # payload
            w3c_data = W3CTraceParent.decode(metadata['traceparent'])
            nr_tracestate = list(W3CTraceState.decode(
                    metadata['tracestate']).values())[0]
            nr_tracestate = NrTraceState.decode(nr_tracestate, None)
            w3c_data.update(nr_tracestate)

            # Remove all trust keys
            decoded['d'].pop('tk', None)
            w3c_data.pop('tk')

            assert decoded['d'] == w3c_data

    _test()


@pytest.mark.parametrize(*_test_matrix)
def test_outbound_payload_outside_transaction(
        mock_grpc_server, method_type, method_name, stub):
    request_type, response_type = method_type.split('_', 1)
    streaming_request = request_type == 'stream'
    streaming_response = response_type == 'stream'
    stub_method = 'DtNoTxn' + method_type.title().replace('_', '')

    request = create_request(streaming_request)

    method_callable = getattr(stub, stub_method)
    method = getattr(method_callable, method_name)

    reply = method(request)

    if isinstance(reply, tuple):
        reply = reply[0]

    try:
        # If the reply was canceled or the server code raises an exception,
        # this will raise an exception which will be recorded by the agent
        if streaming_response:
            reply = list(reply)
        else:
            reply = [reply.result()]
    except (AttributeError, TypeError):
        reply = [reply]

    # Verify there were no DT headers sent
    metadata = json.loads(reply[0].text)

    assert 'newrelic' not in metadata
    assert 'traceparent' not in metadata
    assert 'tracestate' not in metadata



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
from testing_support.fixtures import (
    dt_enabled,
    function_not_called,
    override_application_settings,
    validate_tt_segment_params,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import DatabaseTrace
from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.function_trace import FunctionTrace, function_trace
from newrelic.api.memcache_trace import MemcacheTrace
from newrelic.api.message_trace import MessageTrace
from newrelic.api.solr_trace import SolrTrace
from newrelic.api.time_trace import (
    add_custom_span_attribute,
    current_trace,
    notice_error,
)
from newrelic.api.transaction import current_transaction
from newrelic.common.object_names import callable_name

ERROR = ValueError("whoops")
ERROR_NAME = callable_name(ERROR)


@pytest.mark.parametrize("dt_enabled", (True, False))
@pytest.mark.parametrize("span_events_enabled", (True, False))
@pytest.mark.parametrize("txn_sampled", (True, False))
def test_span_events(dt_enabled, span_events_enabled, txn_sampled):
    guid = "dbb536c53b749e0b"
    sentinel_guid = "0687e0c371ea2c4e"
    function_guid = "482439c52de807ee"
    transaction_name = "OtherTransaction/Function/transaction"
    priority = 0.5

    @function_trace(name="child")
    def child():
        pass

    @function_trace(name="function")
    def function():
        current_trace().guid = function_guid
        child()

    _settings = {"distributed_tracing.enabled": dt_enabled, "span_events.enabled": span_events_enabled}

    count = 0
    if dt_enabled and span_events_enabled and txn_sampled:
        count = 1

    exact_intrinsics_common = {
        "type": "Span",
        "transactionId": guid,
        "sampled": txn_sampled,
        "priority": priority,
        "category": "generic",
    }
    expected_intrinsics = ("timestamp", "duration")

    exact_intrinsics_root = exact_intrinsics_common.copy()
    exact_intrinsics_root["name"] = "Function/transaction"
    exact_intrinsics_root["transaction.name"] = transaction_name
    exact_intrinsics_root["nr.entryPoint"] = True

    exact_intrinsics_function = exact_intrinsics_common.copy()
    exact_intrinsics_function["name"] = "Function/function"
    exact_intrinsics_function["parentId"] = sentinel_guid

    exact_intrinsics_child = exact_intrinsics_common.copy()
    exact_intrinsics_child["name"] = "Function/child"
    exact_intrinsics_child["parentId"] = function_guid

    @validate_span_events(count=count, expected_intrinsics=["nr.entryPoint"])
    @validate_span_events(count=count, exact_intrinsics=exact_intrinsics_root, expected_intrinsics=expected_intrinsics)
    @validate_span_events(
        count=count, exact_intrinsics=exact_intrinsics_function, expected_intrinsics=expected_intrinsics
    )
    @validate_span_events(count=count, exact_intrinsics=exact_intrinsics_child, expected_intrinsics=expected_intrinsics)
    @override_application_settings(_settings)
    @background_task(name="transaction")
    def _test():
        # Force intrinsics
        txn = current_transaction()
        current_trace().guid = sentinel_guid
        txn.guid = guid
        txn._priority = priority
        txn._sampled = txn_sampled

        function()

    _test()


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
    ),
)
def test_each_span_type(trace_type, args):
    @validate_span_events(count=2)
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
        }
    )
    @background_task(name="test_each_span_type")
    def _test():

        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args):
            pass

    _test()


@pytest.mark.parametrize(
    "sql,sql_format,expected",
    (
        pytest.param("a" * 2001, "raw", "".join(["a"] * 1997 + ["..."]), id="truncate"),
        pytest.param("a" * 2000, "raw", "".join(["a"] * 2000), id="no_truncate"),
        pytest.param(
            "select * from %s" % "".join(["?"] * 2000),
            "obfuscated",
            "select * from %s..." % ("".join(["?"] * (2000 - len("select * from ") - 3))),
            id="truncate_obfuscated",
        ),
        pytest.param("select 1", "off", ""),
        pytest.param("select 1", "raw", "select 1"),
        pytest.param("select 1", "obfuscated", "select ?"),
    ),
)
def test_database_db_statement_format(sql, sql_format, expected):
    @validate_span_events(
        count=1,
        exact_agents={
            "db.statement": expected,
        },
    )
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "transaction_tracer.record_sql": sql_format,
        }
    )
    @background_task(name="test_database_db_statement_format")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with DatabaseTrace(sql):
            pass

    _test()


@validate_span_events(
    count=1,
    exact_intrinsics={"category": "datastore"},
    unexpected_agents=["db.statement"],
)
@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
        "span_events.attributes.exclude": ["db.statement"],
    }
)
@background_task(name="test_database_db_statement_exclude")
def test_database_db_statement_exclude():
    transaction = current_transaction()
    transaction._sampled = True

    with DatabaseTrace("select 1"):
        pass


@pytest.mark.parametrize(
    "trace_type,args,attrs",
    (
        (
            DatastoreTrace,
            ("db_product", "db_target", "db_operation"),
            {"db.collection": "db_target", "db.operation": "db_operation"},
        ),
        (
            DatabaseTrace,
            ("select 1 from db_table",),
            {"db.collection": "db_table", "db.statement": "select ? from db_table"},
        ),
    ),
)
def test_datastore_database_trace_attrs(trace_type, args, attrs):
    @validate_span_events(
        count=1,
        exact_agents=attrs,
    )
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
        }
    )
    @background_task(name="test_database_db_statement_exclude")
    def test():
        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args):
            pass

    test()


@pytest.mark.parametrize("exclude_url", (True, False))
def test_external_spans(exclude_url):
    override_settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }

    if exclude_url:
        override_settings["span_events.attributes.exclude"] = ["http.url"]
        exact_agents = {}
        unexpected_agents = ["http.url"]
    else:
        exact_agents = {"http.url": "http://example.com/foo"}
        unexpected_agents = []

    @validate_span_events(
        count=1,
        exact_intrinsics={
            "name": "External/example.com/library/get",
            "type": "Span",
            "sampled": True,
            "category": "http",
            "span.kind": "client",
            "component": "library",
            "http.method": "get",
        },
        exact_agents=exact_agents,
        unexpected_agents=unexpected_agents,
        expected_intrinsics=("priority",),
    )
    @override_application_settings(override_settings)
    @background_task(name="test_external_spans")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with ExternalTrace(library="library", url="http://example.com/foo?secret=123", method="get"):
            pass

    _test()


@pytest.mark.parametrize(
    "kwarg_override,attr_override",
    (
        ({"url": "a" * 256}, {"http.url": "a" * 255}),
        ({"library": "a" * 256}, {"component": "a" * 255}),
        ({"method": "a" * 256}, {"http.method": "a" * 255}),
    ),
)
@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }
)
def test_external_span_limits(kwarg_override, attr_override):

    exact_intrinsics = {
        "type": "Span",
        "sampled": True,
        "category": "http",
        "span.kind": "client",
        "component": "library",
        "http.method": "get",
    }
    exact_agents = {
        "http.url": "http://example.com/foo",
    }
    for attr_name, attr_value in attr_override.items():
        if attr_name in exact_agents:
            exact_agents[attr_name] = attr_value
        else:
            exact_intrinsics[attr_name] = attr_value

    kwargs = {
        "library": "library",
        "url": "http://example.com/foo?secret=123",
        "method": "get",
    }
    kwargs.update(kwarg_override)

    @validate_span_events(
        count=1,
        exact_intrinsics=exact_intrinsics,
        exact_agents=exact_agents,
        expected_intrinsics=("priority",),
    )
    @background_task(name="test_external_spans")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with ExternalTrace(**kwargs):
            pass

    _test()


@pytest.mark.parametrize(
    "kwarg_override,attribute_override",
    (
        ({"host": "a" * 256}, {"peer.hostname": "a" * 255, "peer.address": "a" * 255}),
        ({"port_path_or_id": "a" * 256, "host": "a"}, {"peer.hostname": "a", "peer.address": "a:" + "a" * 253}),
        ({"database_name": "a" * 256}, {"db.instance": "a" * 255}),
    ),
)
@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }
)
def test_datastore_span_limits(kwarg_override, attribute_override):

    exact_intrinsics = {
        "type": "Span",
        "sampled": True,
        "category": "datastore",
        "span.kind": "client",
        "component": "library",
    }

    exact_agents = {
        "db.instance": "db",
        "peer.hostname": "foo",
        "peer.address": "foo:1234",
    }

    for k, v in attribute_override.items():
        if k in exact_agents:
            exact_agents[k] = v
        else:
            exact_intrinsics[k] = v

    kwargs = {
        "product": "library",
        "target": "table",
        "operation": "operation",
        "host": "foo",
        "port_path_or_id": 1234,
        "database_name": "db",
    }
    kwargs.update(kwarg_override)

    @validate_span_events(
        count=1,
        exact_intrinsics=exact_intrinsics,
        expected_intrinsics=("priority",),
        exact_agents=exact_agents,
    )
    @background_task(name="test_external_spans")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with DatastoreTrace(**kwargs):
            pass

    _test()


@pytest.mark.parametrize("collect_span_events", (False, True))
@pytest.mark.parametrize("span_events_enabled", (False, True))
def test_collect_span_events_override(collect_span_events, span_events_enabled):
    spans_expected = collect_span_events and span_events_enabled
    # if collect_span_events and span_events_enabled:
    #     spans_expected = True
    # else:
    #     spans_expected = False

    span_count = 2 if spans_expected else 0

    @validate_span_events(count=span_count)
    @override_application_settings(
        {
            "transaction_tracer.enabled": False,
            "distributed_tracing.enabled": True,
            "span_events.enabled": span_events_enabled,
            "collect_span_events": collect_span_events,
        }
    )
    @background_task(name="test_collect_span_events_override")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with FunctionTrace("span_generator"):
            pass

    if not spans_expected:
        _test = function_not_called("newrelic.core.attribute", "resolve_agent_attributes")(_test)

    _test()


@pytest.mark.parametrize("include_attribues", (True, False))
def test_span_event_agent_attributes(include_attribues):
    override_settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }
    if include_attribues:
        count = 1
        override_settings["attributes.include"] = ["*"]
    else:
        count = 0

    @override_application_settings(override_settings)
    @validate_span_events(count=count, expected_agents=["webfrontend.queue.seconds"])
    @validate_span_events(
        count=count,
        exact_agents={"trace1_a": "foobar", "trace1_b": "barbaz"},
        unexpected_agents=["trace2_a", "trace2_b"],
    )
    @validate_span_events(
        count=count,
        exact_agents={"trace2_a": "foobar", "trace2_b": "barbaz"},
        unexpected_agents=["trace1_a", "trace1_b"],
    )
    @background_task(name="test_span_event_agent_attributes")
    def _test():
        transaction = current_transaction()
        transaction.queue_start = 1.0
        transaction._sampled = True

        with FunctionTrace("trace1") as trace_1:
            trace_1._add_agent_attribute("trace1_a", "foobar")
            trace_1._add_agent_attribute("trace1_b", "barbaz")
            with FunctionTrace("trace2") as trace_2:
                trace_2._add_agent_attribute("trace2_a", "foobar")
                trace_2._add_agent_attribute("trace2_b", "barbaz")

    _test()


class FakeTrace(object):
    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
        (FakeTrace, ()),
    ),
)
@pytest.mark.parametrize("exclude_attributes", (True, False))
def test_span_event_user_attributes(trace_type, args, exclude_attributes):

    _settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }

    forgone_params = [
        "invalid_value",
    ]
    expected_params = {"trace1_a": "foobar", "trace1_b": "barbaz"}
    # We expect user_attributes to be included by default
    if exclude_attributes:
        count = 0
        _settings["attributes.exclude"] = ["*"]
        forgone_params.extend(("trace1_a", "trace1_b"))
        expected_trace_params = {}
    else:
        expected_trace_params = expected_params
        count = 1

    @override_application_settings(_settings)
    @validate_span_events(
        count=count,
        exact_users=expected_params,
        unexpected_users=forgone_params,
    )
    @validate_tt_segment_params(exact_params=expected_trace_params, forgone_params=forgone_params)
    @background_task(name="test_span_event_user_attributes")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args):
            add_custom_span_attribute("trace1_a", "foobar")
            add_custom_span_attribute("trace1_b", "barbaz")
            add_custom_span_attribute("invalid_value", sys.maxsize + 1)

    _test()


@validate_span_events(count=1, exact_users={"foo": "b"})
@dt_enabled
@background_task(name="test_span_user_attribute_overrides_transaction_attribute")
def test_span_user_attribute_overrides_transaction_attribute():
    transaction = current_transaction()

    transaction.add_custom_parameter("foo", "a")
    add_custom_span_attribute("foo", "b")
    transaction.add_custom_parameter("foo", "c")


@override_application_settings({"attributes.include": "*"})
@validate_span_events(count=1, exact_agents={"foo": "b"})
@dt_enabled
@background_task(name="test_span_agent_attribute_overrides_transaction_attribute")
def test_span_agent_attribute_overrides_transaction_attribute():
    transaction = current_transaction()
    trace = current_trace()

    transaction._add_agent_attribute("foo", "a")
    trace._add_agent_attribute("foo", "b")
    transaction._add_agent_attribute("foo", "c")


def test_span_custom_attribute_limit():
    """
    This test validates that span attributes take precedence when
    adding the span and transaction custom parameters that reach the
    maximum user attribute limit.
    """
    span_custom_attrs = []
    txn_custom_attrs = []
    unexpected_txn_attrs = []

    for i in range(128):
        if i < 64:
            span_custom_attrs.append("span_attr%i" % i)
        txn_custom_attrs.append("txn_attr%i" % i)

    unexpected_txn_attrs.extend(span_custom_attrs)
    span_custom_attrs.extend(txn_custom_attrs[:64])
    expected_txn_attrs = {"user": txn_custom_attrs, "agent": [], "intrinsic": []}
    expected_absent_txn_attrs = {"agent": [], "user": unexpected_txn_attrs, "intrinsic": []}

    @override_application_settings({"attributes.include": "*"})
    @validate_transaction_event_attributes(expected_txn_attrs, expected_absent_txn_attrs)
    @validate_span_events(count=1, expected_users=span_custom_attrs, unexpected_users=txn_custom_attrs[64:])
    @dt_enabled
    @background_task(name="test_span_attribute_limit")
    def _test():
        transaction = current_transaction()

        for i in range(128):
            transaction.add_custom_parameter("txn_attr%i" % i, "txnValue")
            if i < 64:
                add_custom_span_attribute("span_attr%i" % i, "spanValue")

    _test()


_span_event_metrics = [("Supportability/SpanEvent/Errors/Dropped", None)]


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
        (FakeTrace, ()),
    ),
)
def test_span_event_error_attributes_notice_error(trace_type, args):

    _settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }

    error = ValueError("whoops")

    exact_agents = {
        "error.class": callable_name(error),
        "error.message": "whoops",
    }

    @override_application_settings(_settings)
    @validate_transaction_metrics(
        "test_span_event_error_attributes_notice_error", background_task=True, rollup_metrics=_span_event_metrics
    )
    @validate_span_events(
        count=1,
        exact_agents=exact_agents,
    )
    @background_task(name="test_span_event_error_attributes_notice_error")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args):
            try:
                raise ValueError("whoops")
            except:
                notice_error()

    _test()


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
    ),
)
def test_span_event_error_attributes_observed(trace_type, args):

    error = ValueError("whoops")

    exact_agents = {
        "error.class": callable_name(error),
        "error.message": "whoops",
    }

    # Verify errors are not recorded since notice_error is not called
    rollups = [("Errors/all", None)] + _span_event_metrics

    @dt_enabled
    @validate_transaction_metrics(
        "test_span_event_error_attributes_observed", background_task=True, rollup_metrics=rollups
    )
    @validate_span_events(
        count=1,
        exact_agents=exact_agents,
    )
    @background_task(name="test_span_event_error_attributes_observed")
    def _test():
        try:
            with trace_type(*args):
                raise error
        except:
            pass

    _test()


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
        (FakeTrace, ()),
    ),
)
@dt_enabled
@validate_span_events(count=1, exact_agents={"error.class": ERROR_NAME, "error.message": "whoops"})
@background_task(name="test_span_event_notice_error_overrides_observed")
def test_span_event_notice_error_overrides_observed(trace_type, args):
    try:
        with trace_type(*args):
            try:
                raise ERROR
            except Exception:
                notice_error()
                raise ValueError  # pylint: disable
    except ValueError:
        pass


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
        (FakeTrace, ()),
    ),
)
@override_application_settings({"error_collector.enabled": False})
@validate_span_events(count=0, expected_agents=["error.class"])
@validate_span_events(count=0, expected_agents=["error.message"])
@dt_enabled
@background_task(name="test_span_event_errors_disabled")
def test_span_event_errors_disabled(trace_type, args):
    with trace_type(*args):
        try:
            raise ValueError("whoops")
        except:
            notice_error()


_metrics = [("Supportability/SpanEvent/Errors/Dropped", 2)]


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (FunctionTrace, ("name",)),
        (FakeTrace, ()),
    ),
)
def test_span_event_multiple_errors(trace_type, args):
    _settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }

    error = ValueError("whoops")

    exact_agents = {
        "error.class": callable_name(error),
        "error.message": "whoops",
        "error.expected": False,
    }

    @override_application_settings(_settings)
    @validate_span_events(
        count=1,
        exact_agents=exact_agents,
    )
    @validate_transaction_metrics("test_span_event_multiple_errors", background_task=True, rollup_metrics=_metrics)
    @background_task(name="test_span_event_multiple_errors")
    def _test():
        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args):
            try:
                raise RuntimeError("whoaa")
            except:
                notice_error()
            try:
                raise RuntimeError("whoo")
            except:
                notice_error()
            try:
                raise ValueError("whoops")
            except:
                notice_error()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import sys

python_version = sys.version_info[:2]

assert python_version in ((2, 7),) or python_version >= (
    3,
    7,
), "The New Relic Python agent only supports Python 2.7 and 3.7+."

with_setuptools = False

try:
    from setuptools import setup

    with_setuptools = True
except ImportError:
    from distutils.core import setup

from distutils.command.build_ext import build_ext  # noqa
from distutils.core import Extension  # noqa
from distutils.errors import (  # noqa
    CCompilerError,
    DistutilsExecError,
    DistutilsPlatformError,
)


def newrelic_agent_guess_next_version(tag_version):
    version, _, _ = str(tag_version).partition("+")
    version_info = list(map(int, version.split(".")))
    if len(version_info) < 3:
        return version
    version_info[1] += 1
    version_info[2] = 0
    return ".".join(map(str, version_info))


def newrelic_agent_next_version(version):
    if version.exact:
        return version.format_with("{tag}")
    else:
        return version.format_next_version(newrelic_agent_guess_next_version, fmt="{guessed}")


script_directory = os.path.dirname(__file__)
if not script_directory:
    script_directory = os.getcwd()

readme_file = os.path.join(script_directory, "README.rst")

if sys.platform == "win32" and python_version > (2, 6):
    build_ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError, IOError)
else:
    build_ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


class BuildExtFailed(Exception):
    pass


class optional_build_ext(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildExtFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except build_ext_errors:
            raise BuildExtFailed()


packages = [
    "newrelic",
    "newrelic.admin",
    "newrelic.api",
    "newrelic.bootstrap",
    "newrelic.common",
    "newrelic.core",
    "newrelic.extras",
    "newrelic.extras.framework_django",
    "newrelic.extras.framework_django.templatetags",
    "newrelic.hooks",
    "newrelic.network",
    "newrelic/packages",
    "newrelic/packages/urllib3",
    "newrelic/packages/urllib3/util",
    "newrelic/packages/urllib3/contrib",
    "newrelic/packages/urllib3/contrib/_securetransport",
    "newrelic/packages/urllib3/packages",
    "newrelic/packages/urllib3/packages/backports",
    "newrelic/packages/urllib3/packages/ssl_match_hostname",
    "newrelic/packages/wrapt",
    "newrelic.samplers",
]

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: System :: Monitoring",
]

kwargs = dict(
    name="newrelic",
    use_scm_version={
        "version_scheme": newrelic_agent_next_version,
        "local_scheme": "no-local-version",
        "git_describe_command": "git describe --dirty --tags --long --match *.*.*",
        "write_to": "newrelic/version.txt",
    },
    setup_requires=["setuptools_scm>=3.2,<7"],
    description="New Relic Python Agent",
    long_description=open(readme_file).read(),
    url="https://docs.newrelic.com/docs/apm/agents/python-agent/",
    project_urls={"Source": "https://github.com/newrelic/newrelic-python-agent"},
    author="New Relic",
    author_email="support@newrelic.com",
    maintainer="New Relic",
    maintainer_email="support@newrelic.com",
    license="Apache-2.0",
    zip_safe=False,
    classifiers=classifiers,
    packages=packages,
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*,!=3.6.*",
    package_data={
        "newrelic": ["newrelic.ini", "version.txt", "packages/urllib3/LICENSE.txt", "common/cacert.pem"],
    },
    extras_require={"infinite-tracing": ["grpcio", "protobuf"]},
)

if with_setuptools:
    kwargs["entry_points"] = {
        "console_scripts": ["newrelic-admin = newrelic.admin:main"],
    }
else:
    kwargs["scripts"] = ["scripts/newrelic-admin"]


def with_librt():
    try:
        if sys.platform.startswith("linux"):
            import ctypes.util

            return ctypes.util.find_library("rt")
    except Exception:
        pass


def run_setup(with_extensions):
    def _run_setup():

        # Create a local copy of kwargs, if there is no c compiler run_setup
        # will need to be re-run, and these arguments can not be present.

        kwargs_tmp = dict(kwargs)

        if with_extensions:
            monotonic_libraries = []
            if with_librt():
                monotonic_libraries = ["rt"]

            kwargs_tmp["ext_modules"] = [
                Extension("newrelic.packages.wrapt._wrappers", ["newrelic/packages/wrapt/_wrappers.c"]),
                Extension(
                    "newrelic.common._monotonic", ["newrelic/common/_monotonic.c"], libraries=monotonic_libraries
                ),
                Extension("newrelic.core._thread_utilization", ["newrelic/core/_thread_utilization.c"]),
            ]
            kwargs_tmp["cmdclass"] = dict(build_ext=optional_build_ext)

        setup(**kwargs_tmp)

    if os.environ.get("TDDIUM") is not None:
        try:
            print("INFO: Running under tddium. Use lock.")
            from lock_file import LockFile
        except ImportError:
            print("ERROR: Cannot import locking mechanism.")
            _run_setup()
        else:
            print("INFO: Attempting to create lock file.")
            with LockFile("setup.lock", wait=True):
                _run_setup()
    else:
        _run_setup()


WARNING = """
WARNING: The optional C extension components of the Python agent could
not be compiled. This can occur where a compiler is not present on the
target system or the Python installation does not have the corresponding
developer package installed. The Python agent will instead be installed
without the extensions. The consequence of this is that although the
Python agent will still run, some non core features of the Python agent,
such as capacity analysis instance busy metrics, will not be available.
Pure Python versions of code supporting some features, rather than the
optimised C versions, will also be used resulting in additional overheads.
"""

with_extensions = os.environ.get("NEW_RELIC_EXTENSIONS", None)
if with_extensions:
    if with_extensions.lower() == "true":
        with_extensions = True
    elif with_extensions.lower() == "false":
        with_extensions = False
    else:
        with_extensions = None

if hasattr(sys, "pypy_version_info"):
    with_extensions = False

if with_extensions is not None:
    run_setup(with_extensions=with_extensions)

else:
    try:
        run_setup(with_extensions=True)

    except BuildExtFailed:

        print(75 * "*")

        print(WARNING)
        print("INFO: Trying to build without extensions.")

        print()
        print(75 * "*")

        run_setup(with_extensions=False)

        print(75 * "*")

        print(WARNING)
        print("INFO: Only pure Python agent was installed.")

        print()
        print(75 * "*")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import (
    function_not_called,
    override_application_settings,
    override_generic_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_count import (
    validate_transaction_count,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.core.config import global_settings


@pytest.mark.parametrize(
    "uri,name,metrics, method_metric",
    (
        # ('/native-simple', '_target_application:NativeSimpleHandler.get', None,
        #         True),
        # ('/simple', '_target_application:SimpleHandler.get', None, True),
        ("/call-simple", "_target_application:CallSimpleHandler.get", None, True),
        ("/super-simple", "_target_application:SuperSimpleHandler.get", None, True),
        ("/coro", "_target_application:CoroHandler.get", None, False),
        ("/fake-coro", "_target_application:FakeCoroHandler.get", None, False),
        ("/coro-throw", "_target_application:CoroThrowHandler.get", None, False),
        ("/init", "_target_application:InitializeHandler.get", None, True),
        ("/multi-trace", "_target_application:MultiTraceHandler.get", [("Function/trace", 2)], True),
    ),
)
@override_application_settings({"attributes.include": ["request.*"]})
def test_server(app, uri, name, metrics, method_metric):
    FRAMEWORK_METRIC = "Python/Framework/Tornado/%s" % app.tornado_version
    METHOD_METRIC = "Function/%s" % name

    metrics = metrics or []
    metrics.append((FRAMEWORK_METRIC, 1))
    metrics.append((METHOD_METRIC, 1 if method_metric else None))

    host = "127.0.0.1:" + str(app.get_http_port())
    namespace, func_name = name.split(".")
    namespace = namespace.replace(":", ".")

    @validate_transaction_metrics(
        name,
        rollup_metrics=metrics,
    )
    @validate_transaction_event_attributes(
        required_params={"agent": ("response.headers.contentType",), "user": (), "intrinsic": ()},
        exact_attrs={
            "agent": {
                "request.headers.contentType": "1234",
                "request.headers.host": host,
                "request.method": "GET",
                "request.uri": uri,
                "response.status": "200",
            },
            "user": {},
            "intrinsic": {"port": app.get_http_port()},
        },
    )
    def _test():
        response = app.fetch(uri, headers=(("Content-Type", "1234"),))
        assert response.code == 200

    if method_metric:
        _test = validate_code_level_metrics(namespace, func_name)(_test)

    _test()


@pytest.mark.parametrize(
    "uri,name,metrics,method_metric",
    (
        ("/native-simple", "_target_application:NativeSimpleHandler.get", None, True),
        ("/simple", "_target_application:SimpleHandler.get", None, True),
        ("/call-simple", "_target_application:CallSimpleHandler.get", None, True),
        ("/super-simple", "_target_application:SuperSimpleHandler.get", None, True),
        ("/coro", "_target_application:CoroHandler.get", None, False),
        ("/fake-coro", "_target_application:FakeCoroHandler.get", None, False),
        ("/coro-throw", "_target_application:CoroThrowHandler.get", None, False),
        ("/init", "_target_application:InitializeHandler.get", None, True),
        ("/ensure-future", "_target_application:EnsureFutureHandler.get", [("Function/trace", None)], True),
        ("/multi-trace", "_target_application:MultiTraceHandler.get", [("Function/trace", 2)], True),
    ),
)
def test_concurrent_inbound_requests(app, uri, name, metrics, method_metric):
    from tornado import gen

    FRAMEWORK_METRIC = "Python/Framework/Tornado/%s" % app.tornado_version
    METHOD_METRIC = "Function/%s" % name

    metrics = metrics or []
    metrics.append((FRAMEWORK_METRIC, 1))
    metrics.append((METHOD_METRIC, 1 if method_metric else None))

    namespace, func_name = name.split(".")
    namespace = namespace.replace(":", ".")

    @validate_transaction_count(2)
    @validate_transaction_metrics(
        name,
        rollup_metrics=metrics,
    )
    def _test():
        url = app.get_url(uri)
        coros = (app.http_client.fetch(url) for _ in range(2))
        responses = app.io_loop.run_sync(lambda: gen.multi(coros))

        for response in responses:
            assert response.code == 200

    if method_metric:
        _test = validate_code_level_metrics(namespace, func_name)(_test)

    _test()


@validate_code_level_metrics("_target_application.CrashHandler", "get")
@validate_transaction_metrics("_target_application:CrashHandler.get")
@validate_transaction_errors(["builtins:ValueError"])
def test_exceptions_are_recorded(app):
    response = app.fetch("/crash")
    assert response.code == 500


@pytest.mark.parametrize(
    "nr_enabled,ignore_status_codes",
    [
        (True, [405]),
        (True, []),
        (False, None),
    ],
)
def test_unsupported_method(app, nr_enabled, ignore_status_codes):
    def _test():
        response = app.fetch("/simple", method="TEAPOT", body=b"", allow_nonstandard_methods=True)
        assert response.code == 405

    if nr_enabled:
        _test = override_ignore_status_codes(ignore_status_codes)(_test)
        _test = validate_transaction_metrics("_target_application:SimpleHandler")(_test)

        if ignore_status_codes:
            _test = validate_transaction_errors(errors=[])(_test)
        else:
            _test = validate_transaction_errors(errors=["tornado.web:HTTPError"])(_test)
    else:
        settings = global_settings()
        _test = override_generic_settings(settings, {"enabled": False})(_test)

    _test()


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("tornado.web:ErrorHandler")
@validate_transaction_event_attributes(
    required_params={"agent": (), "user": (), "intrinsic": ()},
    exact_attrs={
        "agent": {"request.uri": "/does-not-exist"},
        "user": {},
        "intrinsic": {},
    },
)
def test_not_found(app):
    response = app.fetch("/does-not-exist")
    assert response.code == 404


@override_generic_settings(
    global_settings(),
    {
        "enabled": False,
    },
)
@function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
def test_nr_disabled(app):
    response = app.fetch("/simple")
    assert response.code == 200


@pytest.mark.parametrize(
    "uri,name",
    (
        ("/web-socket", "_target_application:WebSocketHandler"),
        ("/call-web-socket", "_target_application:WebNestedHandler"),
    ),
)
def test_web_socket(uri, name, app):
    # import asyncio

    from tornado.websocket import websocket_connect

    namespace, func_name = name.split(":")

    @validate_transaction_metrics(
        name,
        rollup_metrics=[("Function/%s" % name, None)],
    )
    @validate_code_level_metrics(namespace, func_name)
    def _test():
        url = app.get_url(uri).replace("http", "ws")

        async def _connect():
            conn = await websocket_connect(url)
            return conn

        @validate_transaction_metrics(
            name,
        )
        def connect():
            return app.io_loop.run_sync(_connect)

        @function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        def call(call):
            async def _call():
                await conn.write_message("test")
                resp = await conn.read_message()
                assert resp == "hello test"

            app.io_loop.run_sync(_call)

        conn = connect()
        call(conn)
        conn.close()

    _test()


LOOP_TIME_METRICS = (("EventLoop/Wait/" "WebTransaction/Function/_target_application:BlockingHandler.get", 1),)


@pytest.mark.parametrize("yield_before_finish", (True, False))
@validate_transaction_metrics(
    "_target_application:BlockingHandler.get",
    scoped_metrics=LOOP_TIME_METRICS,
)
def test_io_loop_blocking_time(app, yield_before_finish):
    from tornado import gen

    if yield_before_finish:
        url = app.get_url("/block-with-yield/2")
    else:
        url = app.get_url("/block/2")

    coros = (app.http_client.fetch(url) for _ in range(2))
    responses = app.io_loop.run_sync(lambda: gen.multi(coros))

    for response in responses:
        assert response.code == 200
